#include <cmath>
#include <cassert>
#include <cstdio>        // printf
using namespace std;

#include "vector.hpp"
#include "evpiter.hpp"
#include "flags.hpp"

#include "grid.hpp"
#include "grid2d.hpp"
#include "grid3d.hpp"
#include "laplace2d.hpp"
#include "laplace3d.hpp"
#include "mgprocs.hpp"
#include "multigrid.cpp"
#include "evpprocs.hpp"

//--------------------------------------------------------------------------

template<class GRID>
EvpIter<GRID> :: EvpIter(const unsigned int nlevels, const unsigned int nev,
			const bool full_in)
{
 p = nev;
 const CtrlRec simple = {15,1,2,2, bilinear, bilinear,
 			omega_Jacobi, omega_Jacobi, true, false, 1e-04};
// const CtrlRec simple = {15,1,1,1, bilinear,bilinear,GS_forw, GS_backw, true, false, 1e-04};
// const CtrlRec simple = {1,1,1,1, bilinear,bilinear,GS_forw, GS_backw, true, false, 1e-04};
 const GRID coarse(4,true);
// const GRID coarse(3,true);

// mgsolver = new MultiGrid<GRID>(simple,nlevels,coarse);
 mgsolver = new MultiGrid<GRID>(simple,nlevels,coarse,nlevels);
 mgsolver->SetCalcLevel(nlevels);		// Only for Testing here

 N = mgsolver->VectorLength();

 D = new double [N*p];
 W = new double [N*p];
 A = new double [4*p*p];
 B = new double [4*p*p];
 delta_old = new double [2*p];

 //			Addings for full approach in the EVP-solver
 bfull = full_in;
 if (bfull)
  {
    unsigned int i,nn,nx;

    finest_level   = nlevels;
//				Ensure that there are more inner nodes than eigenfunctions to determine
    coarsest_level = 0;
    do
     {
       coarsest_level++;
       mgsolver->SetCalcLevel(coarsest_level);
       nx = mgsolver->GetIntX();
     }
    while (p*2>(nx-1)*(nx-1)*(nx-1));
//    coarsest_level = 1;

    cp = new double* [finest_level+1];
    Xp = new double* [finest_level+1];

    cp[finest_level] = 0;			// use array c in Solve()
    Xp[finest_level] = 0;			// use array X in Solve()
    for ( i = finest_level-1; i >= coarsest_level; i--)
     {
      mgsolver->SetCalcLevel(i);
      nn = mgsolver->VectorLength();
      cp[i] = new double [nn];
      Xp[i] = new double [nn*p];
     }

    mgsolver->SetCalcLevel(finest_level);
  }
 else
  {
   finest_level = coarsest_level = nlevels;
  }

 return;
}

//--------------------------------------------------------------------------

template<class GRID>
EvpIter<GRID> :: ~EvpIter()
{
 delete [] delta_old;
 delete [] B; delete [] A; delete [] W; delete [] D;
 delete mgsolver;

 //			Addings for full approach in the EVP-solver
 if (bfull)
  {
    unsigned int i;
    for ( i = coarsest_level; i <= finest_level-1; i++)
     {
      delete [] Xp[i];  delete [] cp[i];
     }
    delete [] Xp;  delete [] cp;
  }

 return;
}

//--------------------------------------------------------------------------

template<class GRID>
void EvpIter<GRID> :: Solve(const double* c, double* delta, double* X, int errstrat)
//  double c[N], delta[2*p], X[p][N], D[p][N], W[p][N],
//	   A[2*p][2*p], B[2*p][2*p];
{

 if (bfull)
  {
   SolveCoarse2Fine(coarsest_level, finest_level, c, delta, X, errstrat);
  }
 else
  {
   SolveOnLevel(mgsolver->NumAllocatedLevels(), c, delta, X, errstrat);
  }

 return;
}

//--------------------------------------------------------------------------
template<class GRID>
void EvpIter<GRID> :: SolveCoarse2Fine(const unsigned int lev_low, const unsigned int lev_high,
 			const double* c, double* delta, double* X, const int errstrat)
{
  unsigned int i,ll,ip_l,ip_h,nn_l,nn_h;

//					Ensure that the data are available
  if ( lev_high == finest_level)
   {
    cp[lev_high] = (double*) c;
//    cp[lev_high] = const_cast<double*>c;
    Xp[lev_high] = X;				// Use external arrays on highest level
   }
  else
   {
    assert( cp[lev_high] == c );		// Are the right data used on lower levels?
    assert( Xp[lev_high] == X );
   }

//					Restrict  c  and   X   onto the coarser grids
  for ( ll = lev_high; ll >= lev_low+1; ll-- )
   {
    mgsolver->SetCalcLevel(ll-1);nn_l = mgsolver->VectorLength();
    mgsolver->SetCalcLevel(ll);  nn_h = mgsolver->VectorLength();
    mgsolver->Restrict(cp[ll], cp[ll-1]);
//    mgsolver->Inject(cp[ll], cp[ll-1]);

    for ( i = 0; i < NumEigenModes(); i++)
     {
      ip_l = i*nn_l;  ip_h = i*nn_h;
      mgsolver->Restrict(Xp[ll]+ip_h, Xp[ll-1]+ip_l);
     }
   }

//					Start the Full scheme for the EVP-program
  for ( ll = lev_low+1; ll <= lev_high; ll++ )
   {
    SolveOnLevel(ll-1, cp[ll-1], delta, Xp[ll-1], errstrat);
//    assert( errstrat==0 );

    mgsolver->SetCalcLevel(ll-1);  nn_l = mgsolver->VectorLength();
    mgsolver->SetCalcLevel(ll);    nn_h = mgsolver->VectorLength();

    for ( i = 0; i < NumEigenModes(); i++)
     {
      ip_l = i*nn_l;  ip_h = i*nn_h;
      mgsolver->Interpolate(Xp[ll]+ip_h, Xp[ll-1]+ip_l);
     }
   }

 SolveOnLevel(lev_high, cp[lev_high], delta, Xp[lev_high], errstrat);
//    assert( errstrat==0 );

  return;
}
//--------------------------------------------------------------------------

template<class GRID>
void EvpIter<GRID> :: SolveOnLevel(const unsigned int slevel,
				const double* c, double* delta, double* X, int errstrat)
//  double c[N], delta[2*p], X[p][N], D[p][N], W[p][N],
//	   A[2*p][2*p], B[2*p][2*p];
{
 double err, errD, norm_mg;
 unsigned int    i,j,ip,jp,iter;
 int          iter_mg, ierror;

 mgsolver->SetCalcLevel(slevel);
 N = mgsolver->VectorLength();
 mgsolver->AddDiagTerm(c);			// update Helmholtz term
 //NO PresetAB(p, A, B);

 for (i=0; i<p; i++)
   {
     ip = i*N;
     mgsolver->Mult(D+ip,X+ip);			//  d_i := K * x_i
   }

 for (i=0; i<p; i++)
   {					//  delta_i := <x_i,x_i>/<d_i,x_i>
     ip = i*N;
     delta[i] = VDscalar(N, X+ip, X+ip) / VDscalar(N, D+ip, X+ip);
   }
#ifdef OUTPUT_EValues
// for (i=0; i<p; i++) printf("EV0 %d : %6.4e\n",i+1,delta[i]); printf("\n");
#endif

   for (i=0; i<p; i++)
     {
       ip = i*N;
       mgsolver->Mult(D+ip,X+ip);		//   d_i := K * x_i
       VDaXpY(N, D+ip, -delta[i], X+ip);	//   d_i := x_i - delta_i * d_i
     }

 iter = 0;
 do
  {
   iter++;
   VDcopy(p,delta_old,1,delta,1);

   for (i=0; i<p; i++)				//   Solve C w_i = d_i
     {
       ip = i*N;
//       mgsolver->Solve(W+ip, D+ip, iter_mg, norm_mg, false);	// MG
       mgsolver->Solve(W+ip, D+ip, iter_mg, norm_mg);	// MG
//       mgsolver->DiagPC(W+ip, D+ip);			// Diag
//       VDcopy(N,W+ip,1,D+ip,1);			// Identity
  //     cout <<  "  "  << iter_mg;
     }
  // cout << " <-- iter_mg" << endl;

   ResetAB(p, A, B, N, delta, X, D, W);	//   Calculate A, B

//   PrintB(p,A);cout << " #########################" <<endl;
//   PrintB(p,B);cout << " #########################" <<endl;

//  SOLVE EVP  A * Q  =  B * Q * DD
//		Q - matrix 2p*2p, DD - diagmatrix 2p*2p
//		Finde die p gr�ten Eigenwerte in DD
//			und hole entsprechende Eigenvektoren aus Q
   SolveEVP_2(p, A, B, delta, D, N*p, ierror);

//			Stabilization
   if (ierror != 0)
    {
      cout << "I increase entry B(" << ierror << "," << ierror << ")." << endl;
      ResetAB(p, A, B, N, delta, X, D, W);
      for (i=ierror-1; i<2*p; i++)
        {
	 ip = i*(2*p+1);
	 cout << "         B(" << i+1 << "," << i+1 << ") = " << B[ip] << endl;
	 if (B[ip] != 0.0)
             B[ip] = 10.0 *abs(B[ip]);
	 else
             B[ip] = B[ip-2*p-1];
        }
      SolveEVP_2(p, A, B, delta, D, N*p, ierror);
//				Stop if SolveEVP  fails again
      if (ierror!=0)
       { cout << endl << "Sorry, I have to stop the calculation" << endl;
	 exit(-1);
       }
    }

#ifdef OUTPUT_EValues
//  for (i=0; i<p; i++) printf("EV %d : %6.4e\n",i+1,delta[i]); printf("\n");
#endif
 						// Start of update x_i
//		Korrekt ?? Beschleunigung von update ??
   VDcopyC(N*p,D,0.0);				// d_i := 0
   for (i=0; i<p; i++)
   {
     ip = i*N;
     for (j=0; j<p; j++)
     {
//		d_i := d_i + Q_{i,j} * x_j + Q_{2*i,j} * w_j
       jp = j*N;
       VDaXbYZ(N, X+jp, A[i*2*p+j], W+jp, A[i*2*p+p+j], D+ip );
     }
   }
   VDcopy(N*p,X,1,D,1);				// x_i := d_i
 						// End of update x_i
   for (i=0; i<p; i++)
     {
       ip = i*N;
       mgsolver->Mult(D+ip,X+ip);		//   d_i := K * x_i
       VDaXpY(N, D+ip, -delta[i], X+ip);	//   d_i := x_i - delta_i * d_i
     }

  // Abbruchkriterien je nach Eingabeparameter waehlen
  if (errstrat == 1)
  {
    // es sind nur p/2 Eigenvektoren relevant
    err = MaxNormRelativeDiff(p/2, delta, delta_old);
    errD= MaxAbsNorm(N*p/2,D);
  }
  else
  {
    // alle p Eigenvektoren abtesten
    err = MaxNormRelativeDiff(p, delta, delta_old);
    errD= MaxAbsNorm(N*p,D);
  }

#ifdef OUTPUT_Error
//   printf("Iteration %d :   DeltaAccuracy = %6.2e \n",iter,err);
//   printf("             :  DefectAccuracy = %6.2e \n",errD);
//   printf("---------------------------------------------\n\n");
#endif

  }
 while(iter < MAX_ITER_EVP &&
 	(err > ERR_TOLERANCE)); /* --> flags.h	*/
// 	(errD > ERR_TOLERANCE || err > ERR_TOLERANCE)); /* --> flags.h	*/


 printf("\n---------------------------------------------\n");
 if (errstrat == 1)
   printf("Variante 1: p = %d (relevant)\n", p/2);
 else
   printf("Variante 2: p = %d\n", p);
 printf("            l = %d\n\n", mgsolver->GetCalcLevel());

 printf("Iteration %d :   DeltaAccuracy = %6.2e \n",iter,err);
 printf("             :  DefectAccuracy = %6.2e \n",errD);
 printf("---------------------------------------------\n\n");


 return;
}

/* --------------------------------------------------------------------- */
/* 	Matrices A, B have to be symmetrix.				 */
/*	Matrix B has to be positive definte. 				 */
/* --------------------------------------------------------------------- */


template<class GRID>
void EvpIter<GRID> ::
ResetAB(const unsigned int p, double A[],  double B[],
	const unsigned int Nsize, const double delta[], const double X[],
	double D[], const double W[])
{
 unsigned int    i,j,vip,vjp,i11,i12,i21,i22;
 double t0,t1,t2;

//NN				old: normalization
//  for (i=0; i<p; i++)
//  {
//    vip = i*2*p;
//    for (j=0; j<p; j++)
//    {
//      A[vip+j] = 0.0;
//      B[vip+j] = 0.0;
//    }
//    A[vip+i] = delta[i]; 		//	A_11
//    B[vip+i] = 1.0; 			//	B_11
//  }

 for (i=0; i<p; i++)
   {
     vip = i*Nsize;
     i11 = 2*i*p;
     i12 = i11 + p;
     i22 = 2*p*p+(2*i+1)*p;

//				Symmetrie in Blockmatrizen sp�er beachten!!
     for (j=0; j<p; j++)
       {
	 vjp = j*Nsize;
         i21 = 2*p*p + 2*j*p;
	 t0  = VDscalar(Nsize, X+vip, X+vjp);
	 t1  = VDscalar(Nsize, X+vip, W+vjp);
	 t2  = VDscalar(Nsize, W+vip, W+vjp);
//NN
	 A[i11+j] = t0;			//	A_11
         A[i12+j] = t1;			//	A_12
         A[i21+i] = t1;			//	A_21
         A[i22+j] = t2;			//	A_22
       }
   }

 for (i=0; i<p; i++)
   {
     vip = i*Nsize;
     mgsolver->Mult(D+vip,W+vip);		//   d_i := K * w_i
   }

 for (i=0; i<p; i++)
   {
     vip = i*Nsize;
     i11 = 2*i*p;
     i12 = i11 + p;
     i22 = 2*p*p+(2*i+1)*p;
//				Symmetrie in Blockmatrizen sp�er beachten!!
     for (j=0; j<p; j++)
       {
	 vjp = j*Nsize;
         i21 = 2*p*p + 2*j*p;
	 t1  = VDscalar(Nsize, X+vip, D+vjp);
	 t2  = VDscalar(Nsize, W+vip, D+vjp);
         B[i12+j] = t1;			//	B_12
         B[i21+i] = t1;			//	B_21
         B[i22+j] = t2;			//	B_22
       }
   }

 for (i=0; i<p; i++)
   {
     vip = i*Nsize;
     mgsolver->Mult(D+vip,X+vip);		//   d_i := K * x_i
   }
 for (i=0; i<p; i++)
   {
     vip = i*Nsize;
     i11 = 2*i*p;
     for (j=0; j<p; j++)
       {
	 vjp = j*Nsize;
	 t0  = VDscalar(Nsize, X+vip, D+vjp);
//NNNsize
         B[i11+j] = t0;			//	B_11
       }
   }
}
