#ifndef FILE_EVPITER
#define FILE_EVPITER

#include "grid.hpp"
#include "grid2d.hpp"
#include "grid3d.hpp"
#include "laplace2d.hpp"
#include "laplace3d.hpp"
#include "mgprocs.hpp"
#include "multigrid.cpp"
//####################################################################
//		Template class MultiGrid
//####################################################################

/**	The class EvpIter contains the eigenfunction solver for
	grid and operator (Schr�inger equation) provided by the template class GRID.
	The eigenfunction solver is realized via a multigrid preconditioned inverse iteration.
	It is possible to use the FAS (Full approximation Scheme) for
	the EVP.
	@short	Preconditioned inverse iteration for EVP
*/
template<class GRID>
class EvpIter
{
//		Methods
public:
/**	Constructor for the one-level eigenfunction-solver for the operator defined via template GRID..

	The multigrid hierarchie is also constructed herein.
	@param	nlevels		number of levels for the multigrid preconditioner
	@param	nev		number of eigenfunctions to determine
	@param	full_in		switches the FAS (full approximation scheme) approach for the EVP
				to On/Off (true/false)
*/
 EvpIter(const unsigned int nlevels, const unsigned int nev, const bool full_in=false);

 ~EvpIter();

/**	Performs the inverse iteration for solving the EVP.
	General method where the internal variable bfull switches the FAS on/off.
	@param	c		input array[N] of the discrete Potential term (allso Helmholtz term)
	@param	delta		resulting array[2*nev] of the nev lowest eigenvalues - it has to be initialized
	@param	X		resulting array[nev][N] of the nev eigenvectors- it has to be initialized
	@param	errstrat	switch for the accuracy test
		(= 1, i.e., only p/2 eigenfunctions are relevant for that test)
*/
 void Solve(const double* c, double* delta, double* X, int errstrat);

/**	Performs the inverse iteration for solving the EVP on the specified level.
	@param	slevel		specifies the level in the multigrid hierarchy on which
				the EVP has to be solved.
	@see Solve
*/
 void SolveOnLevel(const unsigned int slevel, const double* c, double* delta, double* X, int errstrat);

/**	Performs the inverse iteration for solving the EVP in the FAS way
	starting on the lowest level until the finest level is reached.
	@param	lev_low		specifies the lowest level in the multigrid hierarchy
				where the coarsest EVP has to be solved.
	@param	lev_high	specifies the highest level in the multigrid hierarchy
				where the finest (often, the original) EVP has to be solved.
	@see Solve
*/

 void SolveCoarse2Fine(const unsigned int lev_low, const unsigned int lev_high,
 			const double* c, double* delta, double* X, const int errstrat);

/**	Calculates the dense matrices A and B for the
		small generalized EVP-problem inside the inverse iteration.
	@param	p		number of eigenfunctions to determine
	@param	A		resulting dense matrix[2*p][2*p] (lhs of EVP)
				(see inverse iteration)
	@param	B		resulting dense matrix[2*p][2*p] (rhs of EVP)
				(see inverse iteration)
	@param	Nsize		number of unknowns
	@param	delta		recently calculated eigenvalues[2*p]
	@param	X		recently calculated eigenvectors[p][N]
	@param	D		recently calculated aux. vector[p][N]
				(see inverse iteration)
	@param	W		recently calculated aux. vector[p][N]
				(see inverse iteration)
*/
 void ResetAB(const unsigned int p, double A[],  double B[],
	const unsigned int Nsize, const double delta[],
	const double X[], double D[], const double W[]);

/**
	@return		Number of unknowns..
*/
 int VectorLength() const {return mgsolver->VectorLength();};

/**
	@return		Actual level in the multigrid hierarchy.
*/
 int GetCalcLevel() const {return mgsolver->GetCalcLevel();};

/**
	@return		Number of intervals in x-direction..
*/
 int GetIntX() const {return mgsolver->GetIntX();};

/**
	@return		Number of investigated eigenfunctions.
*/
 int NumEigenModes() const {return p;};

/**
	@return		Space dimension of the problem (2 or 3).
*/
  int  NumberDimensions() const {return mgsolver->NumberDimensions();};

//		Data
private:
  unsigned int p, N;				// #eigenvalues amd #unknowns
  MultiGrid<GRID>  *mgsolver;
  double       *D, *W, *A, *B, *delta_old;	// aux. arrays in inverse iteration

  //			Addings for full approach in the EVP-solver
  bool         bfull;				// FAS approach in EVP-solver
  unsigned int finest_level, coarsest_level;	// index of finest and coarsest level in grid hierarchy
  double       **cp, **Xp;			// Hierarchy of c[] and X[] vectors in the multigrid hierarchy
};

#endif // FILE_EVPITER
