#include <iostream>
#include <cmath>
#include <cassert>
#include <cstdio>        // printf
#include <cstdlib>       // exit
using namespace std;
#include "vector.hpp"
#include "evpprocs.hpp"

// ---------------------------------------------------------------------
void PresetAB(const unsigned int p, double A[], double B[])
{
 int i,j,ip;

 for (i=0; i<p; i++)
   {
     ip = i*2*p;
     for (j=0; j<p; j++)
       {
         A[ip+j] = 0.0;
	 B[ip+j] = 0.0;
       }
     A[ip+i] = 1.0;
     B[ip+i] = 1.0;
   }
}

// ---------------------------------------------------------------------
void SolveEVP_2(const unsigned int p, double A[], double B[], double delta[],
	double aux[], const unsigned int aux_l, int& info)
{
  const char cl = 'L', cv = 'V';
  const int  dim = 2*p, type = 1;
        int  lwork, i;

  lwork = aux_l;

//DSYGV( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK, LWORK, INFO )
  dsygv_(&type, &cv, &cl, &dim, A, &dim, B, &dim, delta, aux, &lwork, &info);

//  cout << " lwork = " << lwork << "  best: " << aux[0] << endl;

  if (info != 0){ cout << "ERROR in dsygv  : " << info << endl;
		  cout << "Probably, matrix B is not positive definite" << endl;
//DB		  PrintB(p,A);
//DB		  PrintB(p,B);
		  if (info > 2*p) {info = info-2*p; return;}
		  assert(0); exit(-1);
		}

//	Eigenfunctions are stored in A now!!
//			reorder such that highest EV (and EF) appears first
  for (i=0; i<p; i++)
  {
    delta[i]  = delta[2*p-1-i];
    VDcopy(2*p, A+i*2*p, 1, A+(2*p-1-i)*2*p, 1);
  }

#ifdef OUTPUT_Efunctions
  for (i=0; i<p; i++)
  {
   printf("EF-2 %d = [ ",i+1);
   for (int j=0; j<2*p; j++)  printf("%6.4e , ",A[i*2*p+j]);
   printf("]\n");
  }
  printf("\n");
#endif

}

// ---------------------------------------------------------------------
void PrintB(const unsigned int p, const double B[])
{
 unsigned int i,j,ip;

 printf("\n             Matrix B\n");
 printf("\n                Matrix B11\n\n");
 for (i=0; i<p; i++)
  {
    ip = i*2*p;
    for (j=0; j<p; j++)
      {
        printf(" %4.2e",B[ip+j]); //printf(" (%d)",ip+j);
      }
    printf("\n");
  }
 printf("\n                Matrix B12\n\n");
 for (i=0; i<p; i++)
  {
    ip = i*2*p+p;
    for (j=0; j<p; j++)
      {
        printf(" %4.2e",B[ip+j]); //printf(" (%d)",ip+j);
      }
    printf("\n");
  }
 printf("\n                Matrix B21\n\n");
 for (i=0; i<p; i++)
  {
    ip = i*2*p+p*2*p;
    for (j=0; j<p; j++)
      {
        printf(" %4.2e",B[ip+j]); //printf(" (%d)",ip+j);
      }
    printf("\n");
  }
 printf("\n                Matrix B22\n\n");
 for (i=0; i<p; i++)
  {
    ip = i*2*p+p*2*p+p;
    for (j=0; j<p; j++)
      {
        printf(" %4.2e",B[ip+j]); //printf(" (%d)",ip+j);
      }
    printf("\n");
  }
}

// ---------------------------------------------------------------------

void PrintEValues(const unsigned int p, const double delta[])
{
 unsigned int i;

 printf("\n##############################################################\n");
 printf("        Eigenvalues\n\n");
 for (i=0; i<p; i++)  printf("EV %d : %10.8e\n",i+1,delta[i]);
 printf("##############################################################\n");
}

//	#######################################################################

// ---------------------------------------------------------------------

void SetEigenmode3D(const unsigned int imode, const unsigned int jmode, const unsigned int kmode,
		const unsigned int n, double x[])
{
 double si,sj,h;
 unsigned int n1,n2,i,j,k,jj;

  n1 = n+1;
  n2 = n1*n1;
  h  = 1.0/n;

  for (i=1; i < n; i++)
  {
    si = 2.0*sqrt(2.0)*sin(imode*M_PI*i*h);
    for (j=1; j < n; j++)
    {
      sj = si*sin(jmode*M_PI*j*h);
      for (k=1; k < n; k++)
      {
	jj = n2*i+n1*j+k;
        *(x+jj) = sj*sin(kmode*M_PI*k*h);
      }
    }
  }
}

// ---------------------------------------------------------------------

void SetFunction3D(const unsigned int ip, const unsigned int n, double x[])
{
 double       h;
 unsigned int n1,n2,i,j,k,jj;

  n1 = n+1;
  n2 = n1*n1;
  h  = 1.0/n1;

  for (i=1; i < n; i++)
  {
    for (j=1; j < n; j++)
    {
      for (k=1; k < n; k++)
      {
	jj = n2*i+n1*j+k;
//        *(x+jj) = i*(n1-i)*j*(n1-j)*k*(n1-k)*h*h*h*powl(i*k*j*h*h*h,ip-1);
        *(x+jj) = i*(n1-i)*j*(n1-j)*k*(n1-k)*h*h*h*pow(i*k*j*h*h*h,ip-1.0);
      }
    }
  }
}

// ---------------------------------------------------------------------
void Init3D_c(const unsigned int nx, double X[])
{
 unsigned int i,j,k,ip;
 double h,t,x;

 h = 1.0/nx;

 ip = 0;
 for (i=0; i<=nx; i++)
 {
   for (j=0; j<=nx; j++)
   {
     for (k=0; k<=nx; k++)
     {
       x = k*h;
       t = (exp(x)+exp(-x))*.5;
       t = 30-30/t/t;
       X[ip] = t;
       ip++;
     }
   }
 }
}

// ---------------------------------------------------------------------
void InitVectors3D(const unsigned int p, const unsigned int nx, double X[])
{
 unsigned int n,m,l,ip,is,js,ks;

 n = nx+1;
 m = nx-1;
 VDcopyC(n*n*n*p,X,0.0);

 for (l=0; l<p; l++)
 {
   is = l/m/m;
   js = (l-is*m*m)/m;
   ks = l-is*m*m-js*m;
   is++; js++; ks++;
//   cout << n << " " << is  << " " << js   << " " << ks << endl;

   ip = l*n*n*n;
   X[ip+is*n*n+js*n+ks] = 1.0;
 }
}

// ---------------------------------------------------------------------
void InitVectors2D(const unsigned int p, const unsigned int nx, double X[])
{
 unsigned int n,m,l,ip,js,ks;

 n = nx+1;
 m = nx-1;
 VDcopyC(n*n*p,X,0.0);

 for (l=0; l<p; l++)
 {
   js = l/m;
   ks = l-js*m;
   js++; ks++;
//   cout << n << " " << js   << " " << ks << endl;

   ip = l*n*n;
   X[ip+js*n+ks] = 1.0;
 }
}

/* --------------------------------------------------------------------- */
void PrintEFunctions3D(const unsigned int p, const unsigned int nx,
							const double X[])
{
 unsigned int n,i,j,k,l,ip;

 n = nx+1;
 printf("\n##############################################################\n");
 printf("        Eigenfunctions\n\n");
 for (l=0; l<p; l++)
 {
   printf("---------------------------------------------------------------\n");
   printf("EF %d :\n",l+1);
   ip = l*n*n*n;
   for (i=0; i<n; i++)
   {
     printf("\n");
     for (j=0; j<n; j++)
     {
       for (k=0; k<n; k++)
       {
         printf("  %6.4e",X[ip]);
	 ip++;
       }
       printf("\n");
     }
   }
 }
 printf("##############################################################\n");
}

// ---------------------------------------------------------------------
void PrintArrayLevel(const unsigned int level,const  unsigned int n,
						const double x[])
{
 unsigned int i,j,ip,offset;

 offset = level*(n+1)*(n+1);
 for (j=0,ip=offset; j<=n; j++)
 {
   for (i=0; i<=n; i++, ip++)
   {
     printf(" %6.4e ",*(x+ip));
   }
   printf("\n ----------- \n");
 }
}

// ---------------------------------------------------------------------

void SolveEVP(unsigned int p, double* A,  double* B, double* delta,
	double* Q, double* aux, unsigned int aux_l)
{
  char cl, ci;
  int  dim, info, type, lwork ,i;

  dim  = 2*p;
  cl   = 'L';
  ci   = 'I';
  type = 1;
  lwork = aux_l;

//DSYGST (ITYPE, UPLO,   N , A, LDA , B, LDB , INFO)

  dsygst_(&type, &cl, &dim, A, &dim, B, &dim, &info);

  if (info != 0)
  		printf("ERROR in dsygst  : %d \n",info);

  lwork = aux_l-2*dim;
  if (lwork <1)
  		printf("ERROR in  SolveEVP : lwork too small %d \n",lwork);

//DSYTRD (UPLO,  N , A, LDA , D    , E  ,     TAU,      WORK,  LWORK, INFO)

  dsytrd_(&cl, &dim, A, &dim, delta, aux, aux+dim, aux+2*dim, &lwork, &info);

  if (info != 0)
  		printf("ERROR in dsytrd  : %d \n",info);

  if (aux_l-dim < 2*dim-2)
  		printf("ERROR in  SolveEVP : aux_l too small %d \n",aux_l);

//DSTEQR(COMPZ,  N ,   D,     E, Z,  LDZ,   WORK , INFO )

  dsteqr_(&ci, &dim, delta, aux, Q, &dim, aux+dim, &info);

  if (info != 0)
  		printf("ERROR in dsteqr  : %d \n",info);

//		reorder such that highest EV (and EF) appears first
  for (i=0; i<2*p; i++) printf("EVV %d : %6.4e\n",i+1,delta[i]); printf("\n");

  for (i=0; i<p; i++)
  {
    delta[i]  = delta[2*p-1-i];
    VDcopy(2*p,Q+i*2*p,1,Q+(2*p-1-i)*2*p,1);
  }

  for (i=0; i<p; i++) printf("EV %d : %6.4e\n",i+1,delta[i]); printf("\n");
}

// ---------------------------------------------------------------------
