#ifndef FILE_GRID
#define FILE_GRID

#include <iostream>
using namespace std;


//GH	Removing of  memory bugs may cause a longer run time.
//GH	Therefore, switch between old ond new via the macro REMOVE_MEM_BUG
//		at least on my laptop (Celeron 1.2MHz) I saw no differrence
//		5%  slower on altix (g++: Itanium2) for LaplaceEqui3D
//		24% slower on altix (icc: Itanium2) for LaplaceEqui3D
// #define REMOVE_MEM_BUG


enum SmoothingType { omega_Jacobi, GS_forw, GS_backw, ADI_forw, ADI_backw };
enum GridtransferType { injection, linear , bilinear };

//	###################################################################
//			Grid
//	###################################################################
/**	The general grid class contains mostly virtual functions
	those provide all the necessary operation for general grids and
	operator handling. This general class usually knows
	only about vector lenght, the next coarser grid and
	the fully occupied coarse grid matrix if the actual grid
	is the coarsest.
	@short The general grid
							*/

class Grid
{
public:
/**	Parameter and default constructor
	@param	n	length of a vector
							*/
  Grid(const int n = 0);

/**	Copy constructor
	@param orig	original grid
							*/
  Grid(const Grid &orig);

/**	Destructor
							*/
  virtual ~Grid();

/**	Assignment operator
	@param orig	original grid
	@return		this grid
							*/
  Grid& operator=(const Grid & orig);

/**
	@return		reference to the next coarser grid
							*/
  virtual const Grid& CoarseGrid()  const {return *coarsegrid_;};


/**	Virtual basis method for injection (restriction)
		UH <-- R * uh

	The Restriction R is contained in the actual derived Grid class.
	@param uh	vector of actual grid
	@param UH	vector of coarser grid
					*/

  virtual void Inject(const double *uh, double *UH) const;

/**	Virtual basis method for restriction
		UH <-- R * uh

	The Restriction R is contained in the actual derived Grid class.
	@param uh	vector of actual grid
	@param UH	vector of coarser grid
						*/
  virtual void Restrict(const double *uh, double *UH) const;

/**	Virtual basis method for interpolation
		wh <-- P * WH

	The Prolongation P is contained in the actual derived Grid class.
	@param wh	vector of actual grid
	@param WH	vector of coarser grid
							*/
  virtual void Interpolate(double *wh, const double *WH) const;

/**	Virtual Basis method for Diagonal preconditioning
		w <-- D^{-1}* r

	The diagonal information is contained in the actual derived Grid class.
	@param w	correction vector
	@param r	defect vector
					*/
  virtual void DiagPC(double *w, const double *r) const;

/**	Virtual Basis method for matrix vector operation
		f <-- K* u

	The matrix information is contained in the actual derived Grid class.
	@param f	resulting vector
	@param u	vector
					*/
  virtual void Mult(double *f, const double *u) const;

/**	Virtual Basis method for defect calculation.
		def <-- f - K* u
	N O T   implemented.

	The matrix information is contained in the actual derived Grid class
	@param f	right hand side vector
	@param u	recent solution vector
	@param def	calculated defect
		 			*/
  virtual void Defect(const double *f, const double *u, double *def) const;

/**	Virtual basis method for combined defect calculation and restriction
		DH <-- R * ( fh - K * uh )

	The matrix information is contained in the actual derived Grid class
	@param fh	right hand side vector on actual grid
	@param uh	recent solution vector on actual grid
	@param DEF	calculated defect projected onto coarse grid
							*/
  virtual void DefRes(const double *fh, const double *uh, double *DH) const;

/**	Virtual basis method for combined interpolation and correction
		uh <-- uh + P * WH

	The Prolongation P is contained in the actual derived Grid class
	@param uh	recent solution vector on actual grid
	@param WH	correction from the coarser grid
							*/
  virtual void IntAdd(double *uh, const double *WH) const;

/**	Virtual basis method for direct solving of (coarse) system.
	@param 	u	solution vector
	@param	f	right hand side vector
							*/
  virtual void DirectSolve(double *u, const double *f) const;

/**	General smoother method.
	@param u	iterative solution vector (in/out)
	@param f	right hand side vector
	@param def	aux. vector, not needed for Gauss-Seidel
	@param type	type of smoother
			(JACOBI/GSFORW/GSBACK,/ADIFORW/ADIBACK = 1/2/3/4/5)
	@param iter	number of smoothing sweeps
							*/
  void Smooth(double *u, const double *f, double *def,
  		const SmoothingType = omega_Jacobi, const int iter=1) const;

/**  	Virtual basis method for Jacobi smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param def	aux. vector
	@param omega	relaxation parameter in Jacobi iteration
	@param iter	number of smoothing sweeps
							*/
  virtual void Jacobi(double *u, const double *f, double *def,
  			const double omega=1.0, const int iter=1) const;

/**	Virtual basis method for Gauss-Seidel forward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param iter	number of smoothing sweeps
							*/
  virtual void GaussSeidelForward(double *u, const double *f, const int iter=1) const;

/**	Virtual basis method for Gauss-Seidel backward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param iter	number of smoothing sweeps
							*/
  virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;

/**	Virtual basis method for ADI forward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param aux	aux. vector
	@param iter	number of smoothing sweeps
							*/
  virtual void ADIForward(double *u, const double *f,
                          double *aux, const int iter=1) const;

/**	Virtual basis method for ADI backward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param aux	aux. vector
	@param iter	number of smoothing sweeps
							*/
  virtual void ADIBackward(double *u, const double *f,
                          double *aux, const int iter=1) const;

/**	Virtual basis method for defect norm calculation
	@param u	recent solution vector
	@param f	right hand side vector
							*/
  virtual double DefectNorm(const double *u, const double *f) const;

/**	Virtual basis method to set all boundary nodes to zero
	@param u	vector from the actual grid
							*/
  virtual void SetBoundaryZero(double *u) const = 0;

/**	Virtual basis method to set all boundary nodes to const
	@param u	vector from the actual grid
	@param c	the constant value
							*/
  virtual void SetBoundaryConst(double *u, const double c) const = 0;

/**	Virtual basis method to add a constant value
	to all boundary nodes
	@param u	vector from the actual grid
	@param c	the constant value
							*/
  virtual void AddBoundaryConst(double *u, const double c) const = 0;

/**	Allocates memory for a grid vector
	@param	u	pointer of the new grid vector
							*/
  void NewGridVector(double* &u) const;

/**	Deallocates memory of a grid vector
	@param	u	pointer of the grid vector
							*/
  void DeleteGridVector(double * &u) const;

/**	Vector length
	@return number of grid points = vector length
							*/
  unsigned int VectorLength() const { return length_; };

/**	Returns  whether actual grid is the coarse grid
	@return is the actual grid the coarsest grid (T/F)
							*/
  bool IsCoarseGrid() const {return coarsegrid_==0 ;};

/**	Space Dimension of the problem
	@return space dimension of the problem ( 2 / 3 )
							*/
  virtual int  NumberDimensions() const {return -1;};

/**	Writes a grid vector to a file (gnuplot grid file)
	@param	u	grid vector
	@param	filename	name of output file
*/
  virtual void VectorToFile(const double u[], const char filename[]) const
    	{cout << "VectorToFile is not implemented" << u << "  " << filename << endl;};

	/**	Calculates the coordinates of the given grid.
	The array @p xccord has to be allocated in advance
	(=VectorLength()*NumberDimensions())!!
	@param	xcoord	2d or 3d array of ccordinates.
*/
  virtual void GetCoords(double xcoord[])  const
  	{cout << "GetCoords is not implemented  " << xcoord << endl;};

protected:

/**	Set number of grids points
	@param number of grid point (= vector lenght)
							*/
  void SetVectorLength(const unsigned int n) { length_ = n; };

/**	Inverts the fully occupied (symmetric) coarse grid matrix.
	Interface to MatrixInversion in vector.cc
	@see vector.hh, vector.cc
		*/
  void InvertMatrix();

private:
   unsigned int length_;		// Number of grid points.

protected:
   const Grid *coarsegrid_;		// coarser level
   double *mat;				// fully occupied matrix
};

#endif // FILE_GRID
