#include <iostream>
#include <cstdlib>
#include <fstream>		// needed for ifstream and ofstream
#include <cmath>
using namespace std;
#include "vector.hpp"
#include "grid2d.hpp"

//	###################################################################
//			Grid2D
//	###################################################################
//	2D tensor product grid

Grid2D :: Grid2D(const int mx, const int my)
 : Grid((mx+1)*(my+1)), nx(mx), ny(my), dimflag(true), dim1(nx+1)
{}
//---------------------------------------------------------------------------

Grid2D :: Grid2D(const int mx)
 : Grid((mx+1)*(mx+1)), nx(mx), ny(mx), dimflag(true), dim1(nx+1)
{}
//---------------------------------------------------------------------------

Grid2D :: ~Grid2D()
{}

//---------------------------------------------------------------------------

void Grid2D :: SetBoundaryZero(double *u) const
{
 SetBoundaryConst(u, 0.0);
}

//---------------------------------------------------------------------------

void Grid2D :: SetBoundaryConst(double *u, const double c) const
{
 unsigned int ii;

 for (ii=0; ii<=nx; ++ii)
   {
      u[idx( 0, ii )] = c;
      u[idx(ny, ii )] = c;
   }
 for (ii=0; ii<=ny; ++ii)
   {
      u[idx(ii, 0 )] = c;
      u[idx(ii,nx )] = c;
   }
}
//---------------------------------------------------------------------------

void Grid2D :: AddBoundaryConst(double *u, const double c) const
{
 unsigned int ii;

 for (ii=0; ii<=nx; ++ii)
   {
      u[idx( 0, ii )] += c;
      u[idx(ny, ii )] += c;
   }
 for (ii=0; ii<=ny; ++ii)
   {
      u[idx(ii, 0 )] += c;
      u[idx(ii,nx )] += c;
   }
}
//---------------------------------------------------------------------------

void Grid2D :: CheckBoundaryZero(const double *u) const
{
 unsigned int ii;

 for (ii=0; ii<=nx; ++ii)
   {
      if ( u[idx( 0, ii )] != 0.0)
      	cout << " an Error at (" << 0 << "," << ii << ")" << endl;
      if ( u[idx(ny, ii )] != 0.0)
      	cout << " an Error at (" << ny << "," << ii << ")" << endl;
   }
 for (ii=0; ii<=ny; ++ii)
   {
      if ( u[idx(ii, 0 )]  != 0.0)
      	cout << " an Error at (" << ii << "," << 0 << ")" << endl;
      if ( u[idx(ii, nx )] != 0.0)
      	cout << " an Error at (" << ii << "," << ny << ")" << endl;
   }
}
//---------------------------------------------------------------------------
//		Dirichlet B.C.

void Grid2D :: Inject(const double *uh, double *UH) const
{

  unsigned int	ncx,ncy,j,jj,kk,ijk,IJK;
  const Grid2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();

  j = 2;
  for (jj=1; jj < ncy; ++jj)
  {
    ijk = idx(j , 2);
    IJK = IDX(jj, 1);
    for (kk=1; kk < ncx; ++kk, ++IJK)
    {
      UH[IJK] = uh[ijk];
      ijk += 2;
    }
    j += 2;
  }
  coarsegrid.SetBoundaryZero(UH);
}
//---------------------------------------------------------------------------:

void Grid2D :: VectorToFile(const double u[], const char filename[]) const
{
				/* coordinates	  */
 double *xcoord = new double [VectorLength()*NumberDimensions()];
 GetCoords(xcoord);

 ofstream outfile(filename);
 if (outfile.fail())
 {
   cout << "File  " << filename << "  does not exist." << endl;
   exit(-1);
 }

 unsigned int ij=0;
 for (unsigned int j=0; j<=GetIntY(); ++j)
 {
 for (unsigned int i=0; i<=GetIntX(); ++i, ++ij)
  {
   for (int k=0; k<NumberDimensions(); ++k)
   {
     outfile << xcoord[ij*NumberDimensions()+k] << "     ";
   }
   outfile << u[ij] << endl;
  }
  outfile << endl;
 }

 delete [] xcoord;
 return;
}

//	###################################################################
//			EquivGrid2D
//	###################################################################
//	Laplace 2D, equidistant tensor product grid

//---------------------------------------------------------------------------

Equi2D :: Equi2D(const int mx, const int my)
 : Grid2D(mx,my), hx(1.0/GetIntX()), hy(1.0/GetIntY())
{}

//---------------------------------------------------------------------------

Equi2D :: Equi2D(const int mx)
 : Grid2D(mx), hx(1.0/GetIntX()), hy(1.0/GetIntX())
{}

//---------------------------------------------------------------------------

Equi2D :: ~Equi2D()
{
}

//---------------------------------------------------------------------------

void Equi2D :: Refine(Equi2D & fine) const
{
  fine = Equi2D(2*nx,2*ny);
  fine.coarsegrid_ = this;
}

//---------------------------------------------------------------------------

void Equi2D :: Restrict(const double *uh, double *UH) const
{
//  const double  CW = 1.;			//  4 because h^2 -> rhs
  const double  CW = 1./4.;			//  normal weight

  double 	t,t2,t4;
  unsigned int	ncx,ncy,nf1,j,jj,kk,ijk,IJK;
  const Equi2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();

//  coarsegrid.SetBoundaryZero(UH);

  for (kk=0; kk <= ncx; ++kk)  UH[kk] = 0.0;	// Dirichlet B.C.
  j = 2;
  for (jj=1; jj < ncy; ++jj)
  {
    ijk = idx(j , 2);
    IJK = IDX(jj, 1);
    UH[IJK-1] = 0.0;				// Dirichlet B.C.
    for (kk=1; kk < ncx; ++kk, ++IJK)
    {
      t  = uh[ijk];

      t2 = ( uh[ijk-nf1] + uh[ijk+nf1] +
	     uh[ijk-  1] + uh[ijk+  1]
	   )/2. ;

      t4 = ( uh[ijk-nf1-1] + uh[ijk-nf1+1] +
	     uh[ijk+nf1-1] + uh[ijk+nf1+1]
	   )/4. ;

      UH[IJK] = (t+t2+t4) * CW ;

      ijk += 2;
    }
    UH[IJK] = 0.0;				// Dirichlet B.C.
    j += 2;
  }
  for (kk=IDX(ncy,0); kk <= IDX(ncy,ncx); ++kk)  UH[kk] = 0.0;	// Dirichlet B.C.

#ifdef DEBUG
    cout << "====================" << endl;
    CheckBoundaryZero(uh);
    cout << "--------------------" << endl;
    coarsegrid.CheckBoundaryZero(UH);
#endif
}

//---------------------------------------------------------------------------

void Equi2D :: Interpolate(double *wh, const double *WH) const
{
 VDcopyC(VectorLength(), wh, 0.0);
 IntAdd(wh,WH);
}

//---------------------------------------------------------------------------

void Equi2D :: IntAdd(double *uh, const double *WH) const
{
  double 	t,t2,t4;
  unsigned int	ncx,ncy,nf1,j,jj,kk,ijk,IJK;
  const Equi2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();

  j = 2;
  for (jj=1; jj < ncy; ++jj)
  {
    ijk = idx(j , 2);
    IJK = IDX(jj, 1);
    for (kk=1; kk < ncx; ++kk, ++IJK)
    {
      t  = WH[IJK];
      t2 = t/2.;
      t4 = t/4.;

      uh[ijk]       += t;

      uh[ijk-nf1]   += t2;
      uh[ijk+nf1]   += t2;
      uh[ijk-1]     += t2;
      uh[ijk+1]     += t2;

      uh[ijk-nf1-1] += t4;
      uh[ijk-nf1+1] += t4;
      uh[ijk+nf1-1] += t4;
      uh[ijk+nf1+1] += t4;

      ijk += 2;
    }
    j += 2;
  }
}

//---------------------------------------------------------------------------:
/**	Calculates the coordinates of the given grid.
	The array @p xccord has to be allocated in advance
	(=VectorLength()*NumberDimensions())!!
	@param	xcoord	2d or 3d array of ccordinates.
*/
 void Equi2D :: GetCoords(double xcoord[]) const
 {
  unsigned int k = 0;
  for (unsigned int j=0; j<=GetIntY(); ++j)
  {
    const double y0 = j*hy;
    for (unsigned int i=0; i<=GetIntX(); ++i, ++k)
    {
      xcoord[2*k+0] = i*hx;
      xcoord[2*k+1] = y0;
    }
  }

  return;
 }

//---------------------------------------------------------------------------

//	###################################################################
//			NonEqui2D
//	###################################################################
/**	An non-equidistant 2D-tensor product grid
							*/
/** 	Parameter constructor having numbers of nodes and
	node coordinates as parameters for all three directions.
							*/
NonEqui2D :: NonEqui2D(
            const int nnx, const double x[],
            const int nny, const double y[])
 : Grid2D(nnx-1,nny-1),
   hhx(new double[GetIntX()+2]), hhy(new double[GetIntY()+2]),
   weight_xl(new double [GetIntX()+2]), weight_xr(new double [GetIntX()+2]),
   weight_yl(new double [GetIntY()+2]), weight_yr(new double [GetIntY()+2])
{
//   hhx = new double[GetIntX()+2]; // Index 1,..,GetIntX() is usefull
//   hhy = new double[GetIntY()+2];
  				// Values at Indices 0 and GetIntX()+1 are set to zero
  hhx[0] = hhx[GetIntX()+1] = 0.0;
  for (unsigned int i=1; i<=GetIntX(); ++i)
  	hhx[i] = x[i]-x[i-1];

  hhy[0] = hhy[GetIntY()+1] = 0.0;
  for (unsigned int i=1; i<=GetIntY(); ++i)
  	hhy[i] = y[i]-y[i-1];

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid

//   weight_xl = new double [GetIntX()+2];
//   weight_xr = new double [GetIntX()+2];
//   weight_yl = new double [GetIntY()+2];
//   weight_yr = new double [GetIntY()+2];

  CalcIntWeights();

  return;
}
//---------------------------------------------------------------------------

NonEqui2D :: NonEqui2D(const int nnx, const double *x)
 : Grid2D(nnx-1,nnx-1),
   hhx(0), hhy(0), weight_xl(0), weight_xr(0), weight_yl(0), weight_yr(0)
{
  if (x==0)
   {
//      cout  << "No hhx-array allocated in NonEqui2D :: NonEqui2D " << endl;
   }
  else
   {
     hhx = new double[GetIntX()+2]; // Index 1,..,GetIntX() is usefull
     hhy = new double[GetIntY()+2];
  				   // Values at Indices 0 and GetIntX()+1 are set to zero
     hhx[0] = hhx[GetIntX()+1] = 0.0;
     hhy[0] = hhy[GetIntY()+1] = 0.0;
     for (unsigned int i=1; i<=GetIntX(); ++i)
  	   hhx[i] = hhy[i] = x[i]-x[i-1];

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid

     weight_xl = new double [GetIntX()+2];
     weight_xr = new double [GetIntX()+2];
     weight_yl = new double [GetIntY()+2];
     weight_yr = new double [GetIntY()+2];

     CalcIntWeights();
   }

  return;
}

//---------------------------------------------------------------------------
/** 	Parameter constructor having numbers of intervals and
	interval lengths as parameters for all three directions.
	xh[i] := x[i] - x[i-1] and
	xh[0:intx-1] is mapped onto hhx[1:intx]
	with  hhx[0] = hx[intx+1] = 0.
*/
NonEqui2D :: NonEqui2D
           (const int intx, const int inty,
            const double xh[], const double yh[])
 : Grid2D(intx,inty),
   hhx(new double[GetIntX()+2]), hhy(new double[GetIntY()+2]),
   weight_xl(new double [GetIntX()+2]), weight_xr(new double [GetIntX()+2]),
   weight_yl(new double [GetIntY()+2]), weight_yr(new double [GetIntY()+2])
{
//   hhx = new double[GetIntX()+2]; // Index 1,..,GetIntX() is usefull
//   hhy = new double[GetIntY()+2];
  				// Values at Indices 0 and GetIntX()+1 are set to zero
  hhx[0] = hhx[GetIntX()+1] = 0.0;
  for (unsigned int i=1; i<=GetIntX(); ++i)
  	hhx[i] = xh[i-1];

  hhy[0] = hhy[GetIntY()+1] = 0.0;
  for (unsigned int i=1; i<=GetIntY(); ++i)
  	hhy[i] = yh[i-1];

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid

//   weight_xl = new double [GetIntX()+2];
//   weight_xr = new double [GetIntX()+2];
//   weight_yl = new double [GetIntY()+2];
//   weight_yr = new double [GetIntY()+2];

  CalcIntWeights();

  return;
}

//---------------------------------------------------------------------------
NonEqui2D :: ~NonEqui2D()
{
  if (hhx!=0) delete [] hhx;
  if (hhy!=0) delete [] hhy;
  if (weight_xl!=0) delete [] weight_xl;
  if (weight_xr!=0) delete [] weight_xr;
  if (weight_yl!=0) delete [] weight_yl;
  if (weight_yr!=0) delete [] weight_yr;
  return;
}

//---------------------------------------------------------------------------
/**	Copy constructor
							*/
NonEqui2D :: NonEqui2D(const NonEqui2D &orig)
 : Grid2D(orig),
    hhx(new double[GetIntX()+2]), hhy(new double[GetIntY()+2]),
   weight_xl(new double [GetIntX()+2]), weight_xr(new double [GetIntX()+2]),
   weight_yl(new double [GetIntY()+2]), weight_yr(new double [GetIntY()+2])
{
//   hhx = new double[GetIntX()+2];
//   hhy = new double[GetIntY()+2];

  for (unsigned int i=0; i<=GetIntX()+1; ++i)
  	hhx[i] = orig.hhx[i];

  for (unsigned int i=0; i<=GetIntY()+1; ++i)
  	hhy[i] = orig.hhy[i];

//   weight_xl = new double [GetIntX()+2];
//   weight_xr = new double [GetIntX()+2];
//   weight_yl = new double [GetIntY()+2];
//   weight_yr = new double [GetIntY()+2];

  for (unsigned int i=0; i<=GetIntX()+1; ++i)
    {
      weight_xl[i] = orig.weight_xl[i];
      weight_xr[i] = orig.weight_xr[i];
    }
  for (unsigned int i=0; i<=GetIntY()+1; ++i)
    {
      weight_yl[i] = orig.weight_yl[i];
      weight_yr[i] = orig.weight_yr[i];
    }

  return;
}

//---------------------------------------------------------------------------

NonEqui2D& NonEqui2D :: operator=(const NonEqui2D & orig)
{
 if ( this != &orig )
 {
   unsigned int i;
   Grid2D::operator=(orig);

   if (hhx!=0) delete [] hhx;
   if (hhy!=0) delete [] hhy;

   hhx = new double[GetIntX()+2];
   hhy = new double[GetIntY()+2];

   for (i=0; i<=GetIntX()+1; ++i)
  	 hhx[i] = orig.hhx[i];

   for (i=0; i<=GetIntY()+1; ++i)
  	 hhy[i] = orig.hhy[i];

   if (weight_xl!=0)  delete [] weight_xl;
   if (weight_xr!=0)  delete [] weight_xr;
   if (weight_yl!=0)  delete [] weight_yl;
   if (weight_yr!=0)  delete [] weight_yr;

   weight_xl = new double [GetIntX()+2];
   weight_xr = new double [GetIntX()+2];
   weight_yl = new double [GetIntY()+2];
   weight_yr = new double [GetIntY()+2];

   for (i=0; i<=GetIntX()+1; ++i)
     {
       weight_xl[i] = orig.weight_xl[i];
       weight_xr[i] = orig.weight_xr[i];
     }
   for (i=0; i<=GetIntY()+1; ++i)
     {
       weight_yl[i] = orig.weight_yl[i];
       weight_yr[i] = orig.weight_yr[i];
     }
 }
 return *this;
}

//---------------------------------------------------------------------------
void NonEqui2D :: CalcIntWeights()
{
  unsigned int i;

  weight_xl[0] = 0.0;
  for (i=1; i < GetIntX()+1; i++)
      weight_xl[i] = hx(i+1)/(hx(i)+hx(i+1));
  weight_xl[GetIntX()+1] = 0.0;

  for (i=0; i < GetIntX()+1; i++)
      weight_xr[i] = hx(i)/(hx(i)+hx(i+1));
  weight_xr[GetIntX()+1] = 0.0;

  weight_yl[0] = 0.0;
  for (i=1; i < GetIntY()+1; i++)
      weight_yl[i] = hy(i+1)/(hy(i)+hy(i+1));
  weight_yl[GetIntY()+1] = 0.0;

  for (i=0; i < GetIntY()+1; i++)
      weight_yr[i] = hy(i)/(hy(i)+hy(i+1));
  weight_yr[GetIntY()+1] = 0.0;

 return;
}

//---------------------------------------------------------------------------

void NonEqui2D :: Interpolate(double *wh, const double *WH) const
{
 VDcopyC(VectorLength(), wh, 0.0);
 IntAdd(wh,WH);
}

//---------------------------------------------------------------------------

void NonEqui2D :: IntAdd(double *uh, const double *WH) const
{
  double 	t;
  unsigned int	nf1,ncx,ncy,j,k,jj,kk,ijk,IJK;
  const NonEqui2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();

  j = 2;
  for (jj=1; jj < ncy; jj++)
  {
    k = 2;
    ijk = idx(j , 2);
    IJK = IDX(jj, 1);
    for (kk=1; kk < ncx; kk++, IJK++)
    {
//cout << k << " # " << wxr(k-1) << "  " << wxl(k+1) << "  " << wyr(j-1) << "  " << wyl(j+1)<<endl;

      t  = WH[IJK];
//      t *= 1.0/200;

      uh[ijk]	+= t;

      uh[ijk-nf1]	+= wyr(j-1)*t;
      uh[ijk+nf1]	+= wyl(j+1)*t;
      uh[ijk-1]		+= wxr(k-1)*t;
      uh[ijk+1]		+= wxl(k+1)*t;

      uh[ijk-nf1-1]	+= wyr(j-1)*wxr(k-1)*t;
      uh[ijk-nf1+1]	+= wyr(j-1)*wxl(k+1)*t;
      uh[ijk+nf1-1]	+= wyl(j+1)*wxr(k-1)*t;
      uh[ijk+nf1+1]	+= wyl(j+1)*wxl(k+1)*t;

      k += 2;  ijk += 2;
    }
    j += 2;
  }
 return;
}

//---------------------------------------------------------------------------

void NonEqui2D :: Restrict(const double *uh, double *UH) const
{
  const double  CW = 1./4.;			//  normal weight
//  const double  CW = 1.0;			//  normal for matrix projection

  double 	t,t2,t4;
  unsigned int	nf1,ncx,ncy,j,k,jj,kk,ijk,IJK;
  const NonEqui2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();

  for (kk=0; kk <= ncx; kk++)  UH[kk] = 0.0;
  j = 2;
  for (jj=1; jj < ncy; jj++)
  {
    k = 2;
    ijk = idx(j , 2);
    IJK = IDX(jj, 1);
    UH[IJK-1]= 0.0;
    for (kk=1; kk < ncx; kk++, IJK++)
    {
//       double tw = 1.0 +
// 	   wyr(j-1) + wyl(j+1) +
// 	   wxr(k-1) + wxl(k+1) +
// 	   wyr(j-1)*( wxr(k-1) + wxl(k+1) ) +
// 	   wyl(j+1)*( wxr(k-1) + wxl(k+1) );

      t  = uh[ijk];

      t2 = wyr(j-1)*uh[ijk-nf1] + wyl(j+1)*uh[ijk+nf1] +
	   wxr(k-1)*uh[ijk-  1] + wxl(k+1)*uh[ijk+  1] ;

      t4 = wyr(j-1)*( wxr(k-1)*uh[ijk-nf1-1]   + wxl(k+1)*uh[ijk-nf1+1]   ) +
	   wyl(j+1)*( wxr(k-1)*uh[ijk+nf1-1]   + wxl(k+1)*uh[ijk+nf1+1]   );

//      UH[IJK] = (t+t2+t4) / tw ;
      UH[IJK] = (t+t2+t4) * CW;

      k += 2;  ijk += 2;
    }
    UH[IJK] = 0.0;
    j += 2;
  }
  for (kk=IDX(ncy,0); kk <= IDX(ncy,ncx); kk++)  UH[kk] = 0.0;

// PrintVector(nf1+1,weight_yl);
// PrintVector(nf1+1,weight_yr);
// cout << " ::::::::::::::::::::::" << endl;
// PrintVector(nf1+1,hhy);

//   CheckBoundaryZero(uh);
//   cout << "--------------------" << endl;
//   coarsegrid.CheckBoundaryZero(UH);
}

//---------------------------------------------------------------------------
//			h[0:n+1], xhf[0:2*n-1]

void NonEqui2D :: Refine1D(const int n, const double h[], double xhf[]) const
{
  int i,ii;
  double tr,tl,p;

  i = 0;
  tr = h[2]/h[1]; tl = 1.0/tr;
  p  = sqrt(tr/tl);
  xhf[i++] = 1.0/(1.0+p)*h[1];
  xhf[i++] = p/(1.0+p)*h[1];
  for (ii=2; ii<n; ii++)
   {
     tl = h[ii-1]/h[ii];
     tr = h[ii+1]/h[ii];
     p  = sqrt(tr/tl);
     xhf[i++] = 1.0/(1.0+p)*h[ii];
     xhf[i++] = p/(1.0+p)*h[ii];
   }
  tl = h[n-1]/h[n];
  tr = 1.0/tl;
  p  = sqrt(tr/tl);
  xhf[i++] = 1.0/(1.0+p)*h[n];
  xhf[i++] = p/(1.0+p)*h[n];

  return;
}

// void NonEqui2D :: Refine1D(const int n, const double h[], double xhf[]) const
// {
//   int i,ii;
//   double tr,tl;
//
//   i = 0;
//   tl = h[1]; tr = h[2];
//   xhf[i++] = tl/(tl+tr)*h[1];
//   xhf[i++] = tr/(tl+tr)*h[1];
//   for (ii=2; ii<n; ii++)
//    {
//      tl = h[ii-1]+h[ii];
//      tr = h[ii]+h[ii+1];
//      xhf[i++] = tl/(tl+tr)*h[ii];
//      xhf[i++] = tr/(tl+tr)*h[ii];
//    }
//   tl = h[n-1];
//   tr = h[n];
//   xhf[i++] = tl/(tl+tr)*h[n];
//   xhf[i++] = tr/(tl+tr)*h[n];
//
//   return;
// }

// void NonEqui2D :: Refine1D(const int n, const double h[], double xhf[]) const
// {
//   int i,ii;
// //  double tr,tl;
//
//   i = 0;
// //  tl = h[1]; tr = h[2];
//   xhf[i++] = 0.5*h[1];
//   xhf[i++] = 0.5*h[1];
//   for (ii=2; ii<n; ii++)
//    {
// //     tl = h[ii-1]+h[ii];
// //     tr = h[ii]+h[ii+1];
//      xhf[i++] = 0.5*h[ii];
//      xhf[i++] = 0.5*h[ii];
//    }
// //  tl = h[n-1];
// //  tr = h[n];
//   xhf[i++] = 0.5*h[n];
//   xhf[i++] = 0.5*h[n];
//
//   return;
// }
//---------------------------------------------------------------------------

void NonEqui2D :: Refine(NonEqui2D & fine) const
{
  int nxc = GetIntX(), nyc =  GetIntY();
  int nxf = 2*nxc,     nyf = 2*nyc;
  double *xhf, *yhf;

  xhf = new double [nxf];
  yhf = new double [nyf];

  Refine1D(nxc, hhx, xhf);
  Refine1D(nyc, hhy, yhf);

  fine = NonEqui2D(nxf,nyf, xhf,yhf);
  fine.coarsegrid_ = this;

  delete [] yhf; delete [] xhf;
}

//---------------------------------------------------------------------------
//			h[0:2*n+1], xhc[0:n-1]

void NonEqui2D :: Coarse1D(const int n, const double h[], double xhf[]) const
{
  int i;

  for (i=0; i<n; i++)
   {
     xhf[i] = h[2*i+1] + h[2*i+2];
   }

  return;
}

//---------------------------------------------------------------------------

bool NonEqui2D :: Coarse(NonEqui2D & coarse)
{
  int nxf = GetIntX(), nyf =  GetIntY();
  int nxc = nxf/2,     nyc = nyf/2;
  double *xhc, *yhc;
  bool bcoarse;

  bcoarse = (2*nxc==nxf && nxc>2) && (2*nyc==nyf && nyc>2);
  if (bcoarse)	// i.e., all coarse dimensions are even
   {

    xhc = new double [nxc];
    yhc = new double [nyc];

    Coarse1D(nxc, hhx, xhc);
    Coarse1D(nyc, hhy, yhc);

    coarse = NonEqui2D(nxc,nyc, xhc,yhc);
    this->coarsegrid_ = &coarse;

    delete [] yhc; delete [] xhc;

//	Calculate Interpolation weights!!
    CalcIntWeights();

#ifdef DEBUG
   cout << endl << nxf << "  " << nxc << endl;
   cout << endl << "HHX : " ;
   for (int i=0; i<nxc+2; i++)
  	  cout << coarse.hhx[i] << "  ";
    cout << endl;
#endif
   }
  else
   {
    this->coarsegrid_ = 0;
   }

  return bcoarse;
}
