#ifndef FILE_GRID2D
#define FILE_GRID2D

#include "grid.hpp"


//	###################################################################
//			Grid2D
//	###################################################################
/**	This general 2D tensor product grid knows, in addition to Grid,
	about the numbers of intervals in both dimensions.
	@short A general 2D tensor product grid
							*/
class Grid2D : public Grid
{
public:
/** 	Parameter constructor
	@param mx	#intervals in x-direction
	@param my	#intervals in y-direction
							*/
  Grid2D(const int mx, const int my);

/** 	Parameter constructor
	@param mx	#intervals in x- and y-direction
							*/
  Grid2D(const int mx=1);

/**	Destructor
*/
  virtual ~Grid2D();

/**
	@return		reference to the next coarser grid
							*/
  const Grid2D& CoarseGrid()  const
  		{return *(Grid2D*)coarsegrid_;};

/**
	@return		#intervalls in x-direction
							*/
  unsigned int GetIntX() const {return nx;};

/**
	@return		#intervalls in y-direction
							*/
  unsigned int GetIntY() const {return ny;};

/**
	@return space dimension of the problem ( here, 2 )
							*/
  int  NumberDimensions() const {return 2;};

/**	Sets all boundary nodes to zero
	@param u	vector from the actual grid
							*/
  virtual void SetBoundaryZero(double *u) const;

/**	Set all boundary nodes to const
	@param u	vector from the actual grid
	@param c	the constant value
							*/
  virtual void SetBoundaryConst(double *u, const double c) const;

/**	Add a constant value to  all boundary nodes
	@param u	vector from the actual grid
	@param c	the constant value
							*/
  virtual void AddBoundaryConst(double *u, const double c) const;

/**	Injection of a vector (the boundary is set to  zero !!)
		UH <-- Inj * uh

	@param uh	vector of actual grid
	@param UH	vector of coarser grid
						*/
  virtual void Inject(const double *uh, double *UH) const;

/**	Writes a grid vector to a file (gnuplot grid file)
	@param	u	grid vector
	@param	filename	name of output file
*/
  void VectorToFile(const double u[], const char filename[]) const;

protected:

/**	First dimension if grid vector is considered as 2d array.
	@return		grid dimension in x-direction (= GetIntX()+1)
					*/
  int Dim1() const {return dim1;};

/**
	@return 1D-index of fine grid point (j,k) with
			j=0,...,GetIntY() and k=0,...,GetIntX()
							*/
  unsigned int idx(const unsigned int  j, const unsigned int k) const
	{return dim1*j+k;};

/**
	@return 1D-index of coarse grid point (j,k) with
		j=0,...,CoarseGrid().GetIntY() and k=0,...,CoarseGrid().GetIntX()
							*/
  unsigned int IDX(const unsigned int  j, const unsigned int k) const
	{return CoarseGrid().dim1*j + k;};

/**	Sets the flag that #intervals are the same in all directions.
	@param	is number of intervals the same ?? (T/F)
							*/
  void SetDimFlag(const bool b) {dimflag = b;};

/**	Checks wether all #intervals are the same in all directions.
	@return	is number of intervals the same ?? (T/F)
							*/
  bool DimsAreEqual() const {return dimflag;};

/**	Checks wether boundary nodes are zero and writes
	error messages in case the test fails.
							*/
  void CheckBoundaryZero(const double *u) const;

protected:
     unsigned int nx,ny;		// #intervals per direction
     bool dimflag;			// same #intervals in all directions?
private:
     int dim1;
};

//	###################################################################
//			Equi2D
//	###################################################################
/**	This equidistant 2D tensor product grid realizes now additionally
	all intergrid transfer routines that don't require the operator.
	@short An equidistant 2D tensor product grid
							*/

class Equi2D : public Grid2D
{
public:
/** 	Parameter constructor
	@param mx	#intervals in x-direction
	@param my	#intervals in y-direction
							*/
  Equi2D(const int mx, const int my);

/** 	Parameter constructor
	@param mx	#intervals in x- and y-direction
							*/
  Equi2D(const int mx=1);

/**	Destructor
*/
  virtual ~Equi2D();

/**	Constructs a finer grid from the given grid in the class.
	@param	the refined new grid
							*/
  void Refine(Equi2D & fine) const;

/**
	@return		reference to the next coarser grid
							*/
 const Equi2D& CoarseGrid()  const
  		{return *(Equi2D*)coarsegrid_;};

/**	Full weighted restriction, i.e., transposed of bilinear interpolation
	with factor  1/4 (boundary is set to  zero !!).
		UH <-- P^T * uh
	@param uh	vector of actual grid
	@param UH	vector of coarser grid
						*/
  virtual void Restrict(const double *uh, double *UH) const;

/**	Bilinear interpolation
		wh <-- P * WH
	@param wh	vector of actual grid
	@param WH	vector of coarser grid
						*/
  virtual void Interpolate(double *wh, const double *WH) const;

/**	Bilinear interpolation and correction
		uh <-- uh + P * WH
	@param uh	recent solution vector on actual grid
	@param WH	correction from the coarser grid
					*/
  virtual void IntAdd(double *uh, const double *WH) const;


//protected:
/**	Calculates the coordinates of the given grid.
	The array @p xcoord has to be allocated in advance
	(=VectorLength()*NumberDimensions())!!
	@param	xcoord	2d or 3d array of ccordinates.
*/
  void GetCoords(double xcoord[]) const;

protected:
  double hx,hy;				// lengths of intervals

};


//	###################################################################
//			NonEqui2D
//	###################################################################
/**	This non-equidistant 2D-tensor product grid realizes
	now additionally all intergrid transfer routines that
	don't require the operator.

	The individual interval lengths in x- (and in y-) direction are
	stored in hhx[i] := x[i] - x[i-1] for hhx[1:GetIntX()]
		and hhx[0]=hhx[GetIntX()+1]=0
	@short	An non-equidistant 2D-tensor product grid
							*/

class NonEqui2D : public Grid2D
{
public:
/** 	Parameter constructor with given coordinate vectors in both directions
	for the tensor product grid (x_i, y_i).
	@param	nnx	#number of nodes in x-direction (=:GetIntX()+1)
	@param	x	vector[0:nnx-1] of x-coordinates
	@param	nny	#number of nodes in y-direction
	@param	y	vector[0:nny-1] of y-coordinates
							*/
  NonEqui2D(const int nnx, const double x[],
            const int nny, const double y[]);

/**	Parameter constructor with the same #nodes
	in all directions (and default constructor).
	@param	nnx	#number of nodes in x-direction (=:GetIntX()+1)
	@param	x	vector[0:nnx-1] of x-coordinates
	*/
  NonEqui2D(const int nnx=1, const double *x = 0);

/** 	Parameter constructor having numbers of intervals and
	interval lengths as parameters for all three directions.

	xh[i] := x[i] - x[i-1] and
	xh[0:intx-1] is mapped onto hhx[1:intx]
	with  hhx[0] = hhx[]= 0.

	@param	intx	#intervals in x-direction
	@param	inty	#intervals in y-direction
	@param	xh	vector[0:intx-1] of interval lengths in x-direction.
			hhx[i] = xn[i-1} for i=1,...,intx and hhx[0]=hhx[intx+1]=0
	@param	yh	vector[0:inty-1] of interval lengths in y-direction
							*/
  NonEqui2D(const int intx, const int inty,
            const double xh[], const double yh[]);

/**	Copy constructor
	@param orig	original grid
							*/
  NonEqui2D(const NonEqui2D &orig);

/**	Destructor
							*/
  virtual ~NonEqui2D();

/**	Assignment operator
	@param orig	original grid
	@return		this grid
							*/
  NonEqui2D& operator=(const NonEqui2D & orig);

/**	Construct a finer grid from the given grid in the class.
	Here several refinement startegies are implemented, see
	grid2d.cc
	@param	the refined new grid
							*/
  void Refine(NonEqui2D & fine) const;

/**	Construct the coarser grid from the actual fine grid
	as long as the number of intervals in all directions
	is an even number (geometrical coarsening).
	Otherwise, the actual grid will marked a coarse grid.
	@param coarse	the coarser grid (or the null pointer)
	@return		is the actual grid already the coarsest grid?? (T/F)

							*/
  bool Coarse(NonEqui2D & coarse);

/**
	@return		reference to the next coarser grid
							*/
  const NonEqui2D& CoarseGrid()  const
  		{return *(NonEqui2D*)coarsegrid_;};

/**	Full weighted restriction, i.e., transposed of bilinear interpolation
	with factor  1/4 (boundary is set to  zero !!).
		UH <-- P^T * uh
	@param uh	vector of actual grid
	@param UH	vector of coarser grid
						*/
  virtual void Restrict(const double *uh, double *UH) const;

/**	Bilinear interpolation
		wh <-- P * WH
	@param wh	vector of actual grid
	@param WH	vector of coarser grid
						*/
  virtual void Interpolate(double *wh, const double *WH) const;

/**	Bilinear interpolation and correction
		uh <-- uh + P * WH
	@param uh	recent solution vector on actual grid
	@param WH	correction from the coarser grid
					*/
  virtual void IntAdd(double *uh, const double *WH) const;

private:
/**	Realizes the refinement in one dimension.
	Several refinement strategies are tried, see grid2d.cc for details.

	This routine could be also defined outside the class.
	@param	n	#nodes on coarse level
	@param	h	vector[0:n+1] of interval lenghts on the coarse level,
				components [1:n] are used
	@param  xfh	vector[0:2*n-1] of new interval lenghts on the finer level,
				all components are used
							*/
  void Refine1D(const int n, const double h[], double xhf[]) const;

/**	Realizes the coarsening in one dimension by merging two
	fine intervals into one coarse interval.
	We assume that all even nodes (0...n) on the fine level
	are also nodes on the coarser level.

	This routine could be also defined outside the class.
	@param	n	#nodes on fine level
	@param	h	vector[0:2*n+1] of interval lenghts on the fine level,
				components [1:2*n] are used
	@param  xfh	vector[0:n-1] of new interval lenghts on the coarserr level,
				all components are used
							*/
  void Coarse1D(const int n, const double h[], double xhc[]) const;

/**	Calculates interpolation weights with respect to the
	non-equidistant discretization.
	The interpolation weights have a similar tensor product
	structure as the intervall lengths.
	But now we have left and right (bottom/top) interpolation
	weights in one point.

	The internal vectors weight_xl, weight_xr, weight_yl and weight_yr
	are calculated in this method.
							*/
  void CalcIntWeights();

//		Data
protected:
  double *hhx, *hhy;			// lengths of intervals

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid
private:
protected:
  double *weight_xl, *weight_xr;	// interpolation weights in x-dir
  double *weight_yl, *weight_yr;	// interpolation weights in y-dir

//		Access methods
protected:
/**	Access to x-intervall lengths on fine grid
	@param	i	interval index [0:GetIntX()+1]
	@return	lenght of x-intervall on fine grid
			(= x[i] - x[i-1] for i in [1:GetIntX()])
							*/
  const double & hx(const int i) const {return hhx[i];}
/**
	@see hx
*/
  const double & hy(const int i) const {return hhy[i];}


/**	Access to x-intervall lengths on coarse grid
	@param	i	interval index [0:CoarseGrid().GetIntX()+1]
	@return	lenght of x-intervall on coarse grid
			(= x[i] - x[i-1] for i in [1:CoarseGrid().GetIntX()])
							*/
  const double & HX(const int i) const {return CoarseGrid().hhx[i];}
/**
	@see HX
*/
  const double & HY(const int i) const {return CoarseGrid().hhy[i];}

/**	Access to interpolation weights in with respect to the
	tensor product structure of the grid data.

	@param	i	index of the node in x-direction [1:GetIntX()]
	@return		weight in x-direction to the left.
  							*/
  const double & wxl(const int i) const {return weight_xl[i];}

/**	Access to interpolation weights in with respect to the
	tensor product structure of the grid data.

	@param	i	index of the node in x-direction [0:GetIntX()]
	@return		weight in x-direction to the right.
  							*/
  const double & wxr(const int i) const {return weight_xr[i];}

/**
	@see wxl
*/
  const double & wyl(const int i) const {return weight_yl[i];}
/**
	@see wxr
*/
  const double & wyr(const int i) const {return weight_yr[i];}
};









#endif
