#ifndef FILE_GRID3D
#define FILE_GRID3D

#include "grid.hpp"

//	###################################################################
//			Grid3D
//	###################################################################
/**	A general 3D tensor product grid
							*/
class Grid3D : public Grid
{
public:
/** 	Parameter constructor
	@param	mx	#intervals in x-direction
	@param	my	#intervals in y-direction
	@param	mz	#intervals in z-direction
							*/
  Grid3D(const int mx, const int my, const int mz);

/**	Parameter constructor with the same #intervals
	in all directions (and default constructor)
	@param	mx	#intervals in x,y,z-direction
*/
  Grid3D(const int mx=1);

  ///
  virtual ~Grid3D();

/**	Reference to the next coarser grid
	@return	next coarser grid
							*/
//  virtual
  const Grid3D& CoarseGrid()  const
  		{return *(Grid3D*)coarsegrid_;};

/**
	@return	#intervalls in x-direction
							*/
  unsigned int GetIntX() const {return nx;};

/**
	@return	#intervalls in y-direction
							*/
  unsigned int GetIntY() const {return ny;};

/**
	@return	#intervalls in z-direction
							*/
  unsigned int GetIntZ() const {return nz;};

/**
	#return	Space Dimension of the problem (=3)
							*/
  int  NumberDimensions() const {return 3;};

/**	Sets all boundary nodes of a vector to zero
	@param	u	I/O array
							*/
  virtual void SetBoundaryZero(double u[]) const;

/**	Sets all boundary nodes of a vector to a constant value
	@param	u	I/O array
	@param	c	constant value
							*/
  virtual void SetBoundaryConst(double u[], const double c) const;

/**	Adds a constant value to all boundary nodes of a vector
	@param	u	I/O array
	@param	c	constant value
							*/
  virtual void AddBoundaryConst(double *u, const double c) const;

/**	Injection of a vector (boundary is set to  zero !!)
	to the next coarser grid
		UH <-- Inj * uh
	@param	uh	fine grid vector
	@param	UH	resulting coarse grid vector
*/
  virtual void Inject(const double uh[], double UH[]) const;

  void  VectorToFile(const double u[], const char filename[]) const;


protected:
/**
	@return	#grid nodes per row
*/
  int Dim1() const {return dim1;};
/**
	@return	#grid nodes per layer
*/
  int Dim2() const {return dim2;};

/**	Index calculation on actual grid
	@param	n2	#grid nodes per layer
	@param	n1	#grid nodes per row
	@param	i	layer index
	@param	j	row index in the layer
	@param	k	node index in the row
	@return	1D-index of actual grid node (i,j,k)
							*/
  unsigned int idx(const unsigned int n2, const unsigned int n1,
        const unsigned int  i, const unsigned int  j, const unsigned int k) const
	{return n2*i+n1*j+k;};

/**	Index calculation on actual grid
	@param	i	layer index
	@param	j	row index in the layer
	@param	k	node index in the row
	@return	1D-index of actual grid node (i,j,k)
							*/
  unsigned int idx(const unsigned int  i, const unsigned int  j, const unsigned int k) const
	{return dim2*i+dim1*j+k;};

/**	Index calculation on next coarser grid
	@param	i	layer index (coarse)
	@param	j	row index in the layer (coarse)
	@param	k	node index in the row (coarse)
	@return	1D-index of next coarser grid node (i,j,k)
							*/
  unsigned int IDX(const unsigned int  i, const unsigned int  j, const unsigned int k) const
	{return CoarseGrid().dim2*i + CoarseGrid().dim1*j + k;};

/**	Sets flag that #intervalls are the same in all directions.
	@param	b	true/false
							*/
  void SetDimFlag(const bool b) {dimflag = b;};

/**	Checks wether all #intervalls are the same in all directions.
	@return	true/false
							*/
  bool DimsAreEqual() const {return dimflag;};

/**	Checks wether boundary nodes are zero.
	If not then messages will be printed.
	@param	u	vector to check.
							*/
  void CheckBoundaryZero(const double u[]) const;

protected:
     unsigned int nx,ny,nz;		// #intervals per direction
     bool dimflag;			// same #intervals in all directions?
private:
     int dim1, dim2;
};


//	###################################################################
//			Equi3D
//	###################################################################
/**	An equidistant 3D-tensor product grid.
	@param	hx	length of invervals in x-direction
	@param	hy	length of invervals in y-direction
	@param	hz	length of invervals in z-direction
*/

class Equi3D : public Grid3D
{
public:
/** 	Parameter constructor
	@param	mx	#intervals in x-direction
	@param	my	#intervals in y-direction
	@param	mz	#intervals in z-direction
							*/
  Equi3D(const int mx, const int my, const int mz);

/**	Parameter constructor with the same #intervals
	in all directions (and default constructor)
	@param	mz	#intervals in z-direction
*/
  Equi3D(const int mx=1);

  ///
  virtual ~Equi3D();

/**	Constructs a finer grid from the actual grid
	@param	fine	the constructed finer grid
							*/
  void Refine(Equi3D & fine) const;

/**	Reference to the next coarser grid
	@return	next coarser grid
							*/
//  virtual
  const Equi3D& CoarseGrid()  const
  		{return *(Equi3D*)coarsegrid_;};

/**	Full weighted Restriction of a vector to the next coarser grid,
	i.e., transposed of the trilinear interpolation
	with factor  1/8 ais used and the boundary is set to  zero.
		UH <-- P^T * uh
	@param	uh	fine grid vector
	@param	UH	resulting coarse grid vector
*/
  virtual void Restrict(const double uh[], double UH[]) const;

/**	Trilinear interpolation from the next coarser grid
		wh <-- P * WH
	@param	wh	resulting fine grid vector
	@param	WH	coarse grid vector
*/
  virtual void Interpolate(double wh[], const double WH[]) const;

/**	Trilinear interpolation from the next coarser grid
	and adding this correction to a vector.
		uh <-- uh + P * WH
	@param	uh	I/O fine grid vector (solution)
	@param	WH	coarse grid vector (correction)
*/
  virtual void IntAdd(double uh[], const double WH[]) const;

/**	Calculates the coordinates of the given grid.
	The array @p xccord has to be allocated in advance
	(=VectorLength()*NumberDimensions())!!
	@param	xcoord	3d array of ccordinates.
*/
  void GetCoords(double xcoord[]) const;

protected:

protected:
  double hx,hy,hz;			// lengths of intervals

};


//	###################################################################
//			NonEqui3D
//	###################################################################
/**	An non-equidistant 3D-tensor product grid
							*/

class NonEqui3D : public Grid3D
{
public:
/** 	Parameter constructor having numbers of nodes and
	node coordinates as parameters for all three directions.
	hx[i] := x[i] - x[i-1] with hx[1:nnx-1] and x[0:nnx-1]
	(hx is stored as hx[0:nnx]).
							*/
  NonEqui3D(const int nnx, const double x[],
            const int nny, const double y[],
	    const int nnz, const double z[]);

/**	Parameter constructor with the same #nodes
	in all directions (and default constructor).	*/
  NonEqui3D(const int nnx=1, const double *x = 0);

/** 	Parameter constructor having numbers of intervals and
	interval lengths as parameters for all three directions.
	xh[i] := x[i] - x[i-1] and
	xh[0:intx-1] is mapped onto hhx[1:intx]
	with  hhx[0] = hx[intx+1] = 0.
							*/
  NonEqui3D(const int intx, const int inty, const int intz,
            const double xh[], const double yh[], const double zh[]);

/**	Copy constructor
							*/
  NonEqui3D(const NonEqui3D &orig);

  ///
  virtual ~NonEqui3D();

  ///
  NonEqui3D& operator=(const NonEqui3D & orig);

/**	Construct a finer grid from the given grid in the class.
							*/
  void Refine(NonEqui3D & fine) const;

/**	Construct the coarser grid in the class from the given fine grid.
							*/
  bool Coarse(NonEqui3D & coarse);

/**	Reference to the next coarser grid
							*/
//  virtual
  const NonEqui3D& CoarseGrid()  const
  		{return *(NonEqui3D*)coarsegrid_;};

/**   Calculates the coordinates wrt. to a unit cube.
*/
  void GetCoords(double xcoord[]) const;

/**	Full weighted Restriction, i.e., transposed of trilinear interpolation
	with factor  1/8 (boundary is set to  zero !!).
		UH <-- P^T * uh				*/
  virtual void Restrict(const double *uh, double *UH) const;

/**	Trilinear interpolation
		wh <-- P * WH				*/
  virtual void Interpolate(double *wh, const double *WH) const;

/**	Trilinear interpolation and correction
		uh <-- uh + P * WH			*/
  virtual void IntAdd(double *uh, const double *WH) const;

private:
/**	Realizes the refinement in one dimension.
							*/
  void Refine1D(const int n, const double h[], double xhf[]) const;
/**	Realizes the refinement in one dimension.
							*/
  void Coarse1D(const int n, const double h[], double xhc[]) const;
/**	Calculates interpolation weights.
							*/
  void CalcIntWeights();

//		Data
protected:
  double *hhx, *hhy, *hhz;		// lengths of intervals

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid
private:
  double *weight_xl, *weight_xr;	// interpolation weights in x-dir
  double *weight_yl, *weight_yr;	// interpolation weights in y-dir
  double *weight_zl, *weight_zr;	// interpolation weights in z-dir

//		Access methods
protected:
/**	Access to hhx[i] of fine grid
							*/
  const double & hx(const int i) const {return hhx[i];}
  const double & hy(const int i) const {return hhy[i];}
  const double & hz(const int i) const {return hhz[i];}

/**	Access to hhx[i] of coarse grid
							*/
  const double & HX(const int i) const {return CoarseGrid().hhx[i];}
  const double & HY(const int i) const {return CoarseGrid().hhy[i];}
  const double & HZ(const int i) const {return CoarseGrid().hhz[i];}

/**	Access to interpolation weights
  							*/
  const double & wxl(const int i) const {return weight_xl[i];}
  const double & wxr(const int i) const {return weight_xr[i];}
  const double & wyl(const int i) const {return weight_yl[i];}
  const double & wyr(const int i) const {return weight_yr[i];}
  const double & wzl(const int i) const {return weight_zl[i];}
  const double & wzr(const int i) const {return weight_zr[i];}
};

#endif // FILE_GRID3D
