#ifndef FILE_LAPLACE2D
#define FILE_LAPLACE2D

#include "grid2d.hpp"

//	###################################################################
//			DiagLaplaceEqui2D
//	###################################################################
/**	Laplace in 2D (provided for -u'' + cu)
		on an equidistant tensor product grid
		using a 5-point stencil from Finite Differences.
		The matrix diagonal is stored as a vector.
		All operations are matrix free.
							*/

class DiagLaplaceEqui2D : public Equi2D
{
public:
/** 	Parameter constructor
							*/
  DiagLaplaceEqui2D(const int mx, const int my, const bool coarse=false);

/**	Parameter constructor with the same #intervals
	in all directions (and default constructor).	*/
  DiagLaplaceEqui2D(const int mx=1, const bool coarse=false);

/**	Copy constructor
							*/
  DiagLaplaceEqui2D(const DiagLaplaceEqui2D &orig);
  ///
  virtual ~DiagLaplaceEqui2D();

  ///
  DiagLaplaceEqui2D& operator=(const DiagLaplaceEqui2D & orig);

/**	Construct a finer grid from the given grid in the class.
							*/
  void Refine(DiagLaplaceEqui2D & fine) const;

/**	Reference to the next coarser grid
							*/
//  virtual
  const DiagLaplaceEqui2D& CoarseGrid()  const
  		{return *(DiagLaplaceEqui2D*)coarsegrid_;};

/**	Basis method for Diagonal preconditioning
		w <-- D^{-1}* r 			*/
  virtual void DiagPC(double *w, const double *r) const
  	{VDdiv(VectorLength(), w, r, diag);};

/**	Basis method for matrix vector operation
		f <-- K* u 			*/
  virtual void Mult(double *f, const double *u) const;

/**	Basis method for defect calculation
		def <-- f - K* u 			*/
  virtual void Defect(const double *f, const double *u, double *def) const;

/**	Combined defect calculation and restriction
		DH <-- R * ( fh - K * uh )		*/
  virtual void DefRes(const double *fh, const double *uh, double *DH) const;

/**	Direct solving of (coarse) system
							*/
  virtual void DirectSolve(double *u, const double *f) const;

/**  	Jacobi smoother
							*/
  virtual void Jacobi(double *u, const double *f, double *def,
  			const double omega=1.0, const int iter=1) const;

/**	Gauss-Seidel forward smoother
							*/
  virtual void GaussSeidelForward(double *u, const double *f, const int iter=1) const;

/**	Gauss-Seidel backward smoother
							*/
  virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;

/**	ADI components in x- and y-direction
							*/
private:
  void ADIx(double u[], const double f[]) const;
  void ADIy(double u[], const double f[]) const;

/**	ADI forward smoother
							*/
public:
  virtual void ADIForward(double *u, const double *f,
                          double *aux, const int iter=1) const;
/**	ADI backward smoother
							*/
  virtual void ADIBackward(double *u, const double *f,
                           double *aux, const int iter=1) const;

/**	Virtual basis method for defect norm calculation
							*/
  virtual double DefectNorm(const double *u, const double *f) const;

/**	Initialization of the operator specific data (const. diagonal).
		The method is used in the constructors, etc..
	If vector cc is set then it will be used for Helmholtz terms
		*/
  void InitOperator(const bool coarse, const double *cc = 0);

  void ProjectOperator(const bool coarse, const DiagLaplaceEqui2D & fine)
  	{};

protected:
/**	Allocates and calculates the fully occupied square matrix.
								*/
  virtual void DeriveMatrix();

/**	Allocates and calculates the matrix diagonal
	@param	cc	vector of the Helmholtz term (=0 if not available)
*/
  virtual void DeriveDiag(const double *cc);

private:
  void FactorADIMatrices();


// private:
protected:
  double bx,by;			// const. off-diagonal entries of matrix
  double *diag;			// const. diagonal entries of matrix
  double *diagx, *diagy;	// ADI: inverse diagonal entries of factored 1D matrix
  double *offx, *offy;		// ADI: off-diagonal entries of factored 1D matrix
};


//	###################################################################
//			DiagLaplaceEqui2Neumann
//	##################################################################
/**	Laplace in 2D (provided for -u'' + cu)
		on an equidistant tensor product grid
		using a 5-point stencil from Finite Differences.
	Homogeneous Neumann boundary conditions
		The matrix diagonal is stored as a vector.
		All operations are matrix free.
							*/

class DiagLaplaceEqui2DNeumann : public DiagLaplaceEqui2D
{
public:
/** 	Parameter constructor
							*/
  DiagLaplaceEqui2DNeumann(const int mx, const int my, const bool coarse=false);

/**	Parameter constructor with the same #intervals
	in all directions (and default constructor).	*/
  DiagLaplaceEqui2DNeumann(const int mx=1, const bool coarse=false);

/**	Copy constructor
							*/
  DiagLaplaceEqui2DNeumann(const DiagLaplaceEqui2DNeumann &orig);
  ///
  virtual ~DiagLaplaceEqui2DNeumann();

/**	Construct a finer grid from the given grid in the class.
							*/
  void Refine(DiagLaplaceEqui2DNeumann & fine) const;

/**	Reference to the next coarser grid
							*/
//  virtual
  const DiagLaplaceEqui2DNeumann& CoarseGrid()  const
  		{return *(DiagLaplaceEqui2DNeumann*)coarsegrid_;};

/**	Basis method for Diagonal preconditioning
		w <-- D^{-1}* r 			*/
  void DiagPC(double *w, const double *r) const
  	{VDdiv(VectorLength(), w, r, diag);};

/**	Basis method for matrix vector operation
		f <-- K* u 			*/
  virtual void Mult(double *f, const double *u) const;
  void Mult_long(double *f, const double *u) const;

 /**	Basis method for defect calculation
 		def <-- f - K* u 			*/
  virtual void Defect(const double *f, const double *u, double *def) const;

 /**	Combined defect calculation and restriction
 		DH <-- R * ( fh - K * uh )		*/
   virtual void DefRes(const double *fh, const double *uh, double *DH) const;

 /**	Direct solving of (coarse) system
 							*/
   virtual void DirectSolve(double *u, const double *f) const;

 /**  	Jacobi smoother
 							*/
//   virtual void Jacobi(double *u, const double *f, double *def,
//   			const double omega=1.0, const int iter=1) const;

 /**	Gauss-Seidel forward smoother
 							*/
//   virtual void GaussSeidelForward(double *u, const double *f, const int iter=1) const;

 /**	Gauss-Seidel backward smoother
 							*/
//   virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;

   void Inject(const double *uh, double *UH) const;
   void Restrict(const double *uh, double *UH) const;
   void IntAdd(double *uh, const double *WH) const;


private:
//   void ADIx(double u[], const double f[]) const;
//   void ADIy(double u[], const double f[]) const;

public:

/**	ADI forward smoother
							*/
//  virtual void ADIForward(double *u, const double *f,
//                          double *aux, const int iter=1) const;
/**	ADI backward smoother
							*/
//  virtual void ADIBackward(double *u, const double *f,
//                           double *aux, const int iter=1) const;

/**	Virtual basis method for defect norm calculation
							*/
  virtual double DefectNorm(const double *u, const double *f) const;

/**	Initialization of the operator specific data (const. diagonal).
  	      The method is used in the constructors, etc..
      If vector cc is set then it will be used for Helmholtz terms
 							*/
//  void InitOperator(const bool coarse, const double *cc = 0);

  static int count_defect;
protected:

/**	Allocates and calculates the fully occupied square matrix.
								*/
  void DeriveMatrix();

/**	Allocates and calculates the matrix diagonal
	@param	cc	vector of the Helmholtz term (=0 if not available)
*/
  void DeriveDiag(const double *cc);

//  void FactorADIMatrices();

};

//	###################################################################
//			LaplaceNonEqui2D
//	###################################################################
/**	Laplace in 2D (provided for -u'' + c )
		on a non-equidistant tensor product grid
		using a 5-point stencil from Finite Differences.
		The matrix diagonal is stored as a vector.
		The  off-diagonals are stored as smaller vectors.
		All operations are matrix free.

	The individual interval lengths in x- (and in y-) direction are
	stored in hhx[i] := x[i] - x[i-1] for hhx[1:GetIntX()]
		and hhx[0]=hhx[GetIntX()+1]=0
	@short 2D Laplace on a non-equidistant tensor product grid
							*/

class LaplaceNonEqui2D : public NonEqui2D
{
public:
/** 	Parameter constructor with given coordinate vectors in both directions
	for the tensor product grid (x_i, y_i).
	@param	nnx	#number of nodes in x-direction (=:GetIntX()+1)
	@param	x	vector[0:nnx-1] of x-coordinates
	@param	nny	#number of nodes in y-direction
	@param	y	vector[0:nny-1] of y-coordinates
	@param	coarse	Is this the coarse grid? (T/F)
							*/
  LaplaceNonEqui2D(const int nnx, const double x[],
                   const int nny, const double y[],
	           const bool coarse=false);

/**	Parameter constructor with the same #nodes
	in all directions (and default constructor).
	@param	nnx	#number of nodes in x-direction (=:GetIntX()+1)
	@param	x	vector[0:nnx-1] of x-coordinates
	@param	coarse	Is this the coarse grid? (T/F)
	*/
  LaplaceNonEqui2D(const int nnx=1, const double *x = 0, const bool coarse=false);

/** 	Parameter constructor having numbers of intervals and
	interval lengths as parameters for all three directions.

	xh[i] := x[i] - x[i-1] and
	xh[0:intx-1] is mapped onto hhx[1:intx]
	with  hhx[0] = hhx[]= 0.

	@param	intx	#intervals in x-direction
	@param	inty	#intervals in y-direction
	@param	xh	vector[0:intx-1] of interval lengths in x-direction.
			hhx[i] = xn[i-1} for i=1,...,intx and hhx[0]=hhx[intx+1]=0
	@param	yh	vector[0:inty-1] of interval lengths in y-direction
							*/
  LaplaceNonEqui2D(const int intx, const int inty,
            const double xh[], const double yh[],
	    const bool coarse=false);

/**	Copy constructor
	@param orig	original grid
							*/
  LaplaceNonEqui2D(const LaplaceNonEqui2D &orig);

/**	Destructor
							*/
  virtual ~LaplaceNonEqui2D();

/**	Assignment operator
	@param orig	original grid
	@return		this grid
							*/
  LaplaceNonEqui2D& operator=(const LaplaceNonEqui2D & orig);

/**	Construct a finer grid from the given grid in the class.
	Here several refinement startegies are implemented, see
	grid2d.cc
	@param	the refined new grid
	@see grid2d.hh, grid2d.cc
							*/
  void Refine(LaplaceNonEqui2D & fine) const;

/**
	@return		reference to the next coarser grid
							*/
  const LaplaceNonEqui2D& CoarseGrid()  const
  		{return *(LaplaceNonEqui2D*)coarsegrid_;};

/**	Diagonal preconditioning
		w <-- D^{-1}* r

	The matrix information is contained in the class.
	@param	w	correctiojn vector
	@param	r	defect vector
					*/
  virtual void DiagPC(double *w, const double *r) const
  	{VDdiv(VectorLength(), w, r, diag);};

/**	Matrix vector operation
		f <-- K* u

	The matrix information is contained in the actual derived Grid class.
	@param f	resulting vector
	@param u	vector
					*/
  virtual void Mult(double *f, const double *u) const;

/**	Defect calculation
		def <-- f - K* u

	The matrix information is contained in the actual derived Grid class
	@param f	right hand side vector
	@param u	recent solution vector
	@param def	calculated defect
		 			*/
  virtual void Defect(const double *f, const double *u, double *def) const;

/**	Combined defect calculation and restriction
		DH <-- R * ( fh - K * uh )

	The matrix information is contained in the actual derived Grid class
	@param fh	right hand side vector on actual grid
	@param uh	recent solution vector on actual grid
	@param DEF	calculated defect projected onto coarse grid
							*/
  virtual void DefRes(const double *fh, const double *uh, double *DH) const;

/**	Direct solving of (coarse) system.
	@param 	u	solution vector
	@param	f	right hand side vector
							*/
  virtual void DirectSolve(double *u, const double *f) const;

/**  	Jacobi smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param def	aux. vector
	@param omega	relaxation parameter in Jacobi iteration
	@param iter	number of smoothing sweeps
							*/
  virtual void Jacobi(double *u, const double *f, double *def,
  			const double omega=1.0, const int iter=1) const;

/**	Gauss-Seidel forward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param iter	number of smoothing sweeps
							*/
  virtual void GaussSeidelForward(double *u, const double *f, const int iter=1) const;

/**	Gauss-Seidel backward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param iter	number of smoothing sweeps
							*/
  virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;


private:

/**	ADI iteration in x-direction for (Laplace + c) and
			hom. Dirichlet B.C.

	The operator is
	rho*I = sqrt(2)*Laplace_y + c/2 *I

	@param u	iterative solution vector
	@param f	right hand side vector
							*/
  void ADIx(double u[], const double f[]) const;

/**	ADI iteration in y-direction for (Laplace + c) and
			hom. Dirichlet B.C.

	The operator is
	rho*I = sqrt(2)*Laplace_x + c/2 *I

	@param u	iterative solution vector
	@param f	right hand side vector
							*/
  void ADIy(double u[], const double f[]) const;

public:
/**	ADI forward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param aux	aux. vector
	@param iter	number of smoothing sweeps
							*/
  virtual void ADIForward(double *u, const double *f,
                          double *aux, const int iter=1) const;

/**	ADI backward smoother
	@param u	iterative solution vector(in/out)
	@param f	right hand side vector
	@param aux	aux. vector
	@param iter	number of smoothing sweeps
							*/
  virtual void ADIBackward(double *u, const double *f,
                           double *aux, const int iter=1) const;

/**	Virtual basis method for defect norm calculation
	@param u	recent solution vector
	@param f	right hand side vector
							*/
  virtual double DefectNorm(const double *u, const double *f) const;

/**	Initialization of the operator specific data (const. diagonal).
		The method is used in the constructors, etc..
	If vector cc is set then it will be used for Helmholtz term c
	in ( -u'' + c )
	@param	coarse	is this the coarse grid? (T/F)
	@param	cc	vector values of c
		*/
  void InitOperator(const bool coarse, const double *cc = 0);

/**	Initialization of the actual grid operator by
	projecting the the fine grid operator onto the actual coarser grid.
	@param	coarse	is this already the coarse grid? (T/F)
	@param	fine
*/
  void ProjectOperator(const bool coarse, const LaplaceNonEqui2D & fine);

protected:
/**	Allocates and calculates the fully occupied quadratic matrix.
								*/
  void DeriveMatrix();

/**	Allocates and calculates the matrix diagonal
	@param	cc	vector of the Helmholtz term (=0 if not available)
*/
  void DeriveDiag(const double *cc);

private:
  void OffDiag1D(const int n, const double h[], double bl[], double br[]);

  void ProjectOffDiag1D(const int nnc, double cbl[], double cbr[],
  			const double fbl[], const double fbr[],
			const double wl[], const double wr[] );


  void CalcIntWeights1D(const int nnodes, const double bl[], const double br[],
			double wl[], double wr[] );
public:
  void CalcIntWeights();


private:
  double *diag;			// diagonal entries of matrix
  double *bxr, *bxl;		// off-diagonal entries of matrix in x-direction
  double *byr, *byl;
  double *diaga;	// ADI: inverse diagonal entries of factored 1D matrix
  double *offa;		// ADI: off-diagonal entries of factored 1D matrix
};


#endif
