#include <iostream>
#include <cstdlib>
#include <cmath>
using namespace std;
#include "vector.hpp"
#include "grid3d.hpp"
#include "laplace3d.hpp"

//#define DEBUG_MATRIX

//	###################################################################
//			DiagLaplaceEqui3D
//	###################################################################
//	Laplace 3D, equidistant tensor product grid

//---------------------------------------------------------------------------

DiagLaplaceEqui3D ::
DiagLaplaceEqui3D(const int mx, const int my, const int mz, const bool coarse)
 : Equi3D(mx,my,mz), diag(0), bx(0), by(0), bz(0)
{
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui3D :: DiagLaplaceEqui3D(const int mx, const bool coarse)
 : Equi3D(mx,mx,mx), diag(0), bx(0), by(0), bz(0)
{
  InitOperator(coarse);}

//---------------------------------------------------------------------------

DiagLaplaceEqui3D :: ~DiagLaplaceEqui3D()
{
 DeleteGridVector(diag);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui3D :: DiagLaplaceEqui3D(const DiagLaplaceEqui3D &orig)
: Equi3D(orig), diag(0), bx(orig.bx), by(orig.by), bz(orig.bz)
{
 NewGridVector(diag);
 VDcopy( orig.VectorLength(), diag, 1, orig.diag, 1 );
}

//---------------------------------------------------------------------------
DiagLaplaceEqui3D& DiagLaplaceEqui3D :: operator=(const DiagLaplaceEqui3D & orig)
{
 if ( this != &orig )
 {
   Equi3D::operator=(orig);

   bx = orig.bx;  by = orig.by;  bz = orig.bz;
   DeleteGridVector(diag);
   NewGridVector(diag);
   VDcopy( orig.VectorLength(), diag, 1, orig.diag, 1 );
 }
 return *this;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3D ::InitOperator(const bool coarse, const double *cc)
{
  bx = -1.0/hx/hx;  by = -1.0/hy/hy;  bz = -1.0/hz/hz;

  DeriveDiag(cc);	//	bx,by,bz are used in diag calculation

  if (coarse)
    {			//	diag is used in matrix calculation
      DeriveMatrix();
      InvertMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Diagonal" << endl;
      PrintMatrix(VectorLength(), mat);
#endif
    }
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3D :: Refine(DiagLaplaceEqui3D & fine) const
{
//						not a coarse grid
  fine = DiagLaplaceEqui3D(2*nx,2*ny,2*nz,false);
  fine.coarsegrid_ = this;
//  fine.DeriveMatrix()  must not be executed since this is already a fine grid ;
}

//---------------------------------------------------------------------------
//	Derive matrix diagonal for the 7 point stencil
//		and Dirichlet boundary conditions

void DiagLaplaceEqui3D :: DeriveDiag(const double *cc)
{
 int n=VectorLength();

 if (diag==0)   NewGridVector(diag);
 VDcopyC( n, diag, -2.0*(bx+by+bz) );
 if  (cc !=0 ) 			//	Add Helmholtz  term
       VDadd( n, diag, cc );
 SetBoundaryConst(diag, 1.0); 	//	Dirichlet boundary

#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix" << endl;
  PrintVector(n,diag);
#endif
}
//---------------------------------------------------------------------------
//	Derive matrix for the 7 point stencil
//		and Dirichlet boundary conditions

void DiagLaplaceEqui3D :: DeriveMatrix()
{
 unsigned int n1,n2,ijk,id;
 unsigned int nn = VectorLength();
 unsigned long int nn2 = nn*nn;

 n1 = nx+1;  n2 = (nx+1)*(ny+1);

//	Allocate memory for fully occupied matrix if called the first time
//			(Assumes no changes in dimensions!!)
 if (mat==0)
  {
    mat = new double[nn2];
  }
 VDcopyC(nn2, mat, 0.0);	// set matrix to Zero
 for ( unsigned int i=0; i<nn; ++i )  mat[i*(nn+1)] = 1.0;	// Initialize diagonal with 1

 for ( unsigned int i=1; i<nz; ++i )
  {
   for ( unsigned int j=1; j<ny; ++j )
    {
     for ( unsigned int k=1; k<nx; ++k )
      {
        ijk = idx(n2,n1,i,j, k);
	id  = ijk*nn + ijk;
	mat[id] = diag[ijk];
	if ( k != 1    ) mat[id-1 ] = bx;
	if ( k != nx-1 ) mat[id+1 ] = bx;
	if ( j != 1    ) mat[id-n1] = by;
 	if ( j != ny-1 ) mat[id+n1] = by;
	if ( i != 1    ) mat[id-n2] = bz;
 	if ( i != nz-1 ) mat[id+n2] = bz;
     }
    }
  }
#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix" << endl;
  PrintMatrix(nn,mat);
#endif
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3D :: DirectSolve(double u[], const double f[]) const
{
 if ( nx != 2 ||  ny != 2 || nz != 2)
  {
   Equi3D::DirectSolve(u,f);
  }
 else
  {
   u[13] = f[13] / diag[13];
  }
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3D :: DefRes(const double fh[], const double uh[],
				 double DH[]) const
{
//  const double  CW = 4./8.;			//  4 because h^2 -> rhs
  const double  CW = 1./8.;			//  normal weight

  double 	t,t1,t2a,t2b,t2c,t4a,t4b,t4c,t8;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
#ifdef REMOVE_MEM_BUG
          bool  bi,bj,bk;
#endif
  const DiagLaplaceEqui3D   &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = nx+1;    nf2 = (nx+1)*(ny+1);
  nc1 = ncx+1;   nc2 = (ncx+1)*(ncy+1);

  VDcopyC(coarsegrid.VectorLength(), DH, 0.0e0);

  i = 2;
  for (ii=1; ii <= ncz; ++ii)
  {
#ifdef REMOVE_MEM_BUG
   bi = ( i !=nz );
#endif
   j = 2;
   for (jj=1; jj <= ncy; ++jj)
    {
#ifdef REMOVE_MEM_BUG
      bj = ( j != ny);
#endif
      k = 2;
      IJK = idx(nc2,nc1,ii,jj, 1);
      for (kk=1; kk <= ncx; ++kk, ++IJK)
      {
//			Compiler besser untersttzen!!
//
#ifdef REMOVE_MEM_BUG
        bk = ( k != nx );
#endif
	ijk = idx(nf2,nf1,i-1,j-1,k-1);
        t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
        t8  = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 8.0 * CW;

	ijk = idx(nf2,nf1,i-1,j-1,k);
#ifdef REMOVE_MEM_BUG
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + ( bk ? uh[ijk+1] : 0.0)  );
#else
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#endif
	t4a = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 4.0 * CW;
//h	if (k==nx)  t4a=0.0;		// inefficient

	ijk = idx(nf2,nf1,i-1,j,k-1);
#ifdef REMOVE_MEM_BUG
        t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + ( bj ? uh[ijk+nf1] : 0.0)  ) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#else
        t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#endif
	t4b = ( fh[ijk] - diag[ijk] * uh[ijk] - t )  / 4.0 * CW;
//h	if (j==ny)  t4b=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j-1,k-1);
#ifdef REMOVE_MEM_BUG
	t   = bz*(uh[ijk-nf2] + ( bi ? uh[ijk+nf2] :0.0 )  ) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#else
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#endif
	t4c = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 4.0 * CW;
//h	if (i==nz)  t4c=0.0;		// inefficient

	ijk = idx(nf2,nf1,i-1,j,k);
#ifdef REMOVE_MEM_BUG
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + ( bj ? uh[ijk+nf1] : 0.0)  ) +
	      bx*(uh[ijk-1]   + ( bk ? uh[ijk+1] : 0.0)  );
#else
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#endif

	t2a = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 2.0 * CW;
//h	if (k==nx || j==ny)  t2a=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j-1,k);
#ifdef REMOVE_MEM_BUG
	t   = bz*(uh[ijk-nf2] + ( bi ? uh[ijk+nf2] :0.0 )  ) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + ( bk ? uh[ijk+1] : 0.0)  );
#else
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#endif
	t2b = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 2.0 * CW;
//h	if (i==nz || k==nx)  t2b=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j,k-1);
#ifdef REMOVE_MEM_BUG
	t   = bz*(uh[ijk-nf2] + ( bi ? uh[ijk+nf2] :0.0 )  ) +
	      by*(uh[ijk-nf1] + ( bj ? uh[ijk+nf1] : 0.0)  ) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#else
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#endif
        t2c = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 2.0 * CW;
//h	if (i==nz || j==ny)  t2c=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j,k);
#ifdef REMOVE_MEM_BUG
        t   = bz*(uh[ijk-nf2] + ( bi ? uh[ijk+nf2] :0.0 )  ) +
	      by*(uh[ijk-nf1] + ( bj ? uh[ijk+nf1] : 0.0)  ) +
	      bx*(uh[ijk-1]   + ( bk ? uh[ijk+1] : 0.0)  );
#else
	t   = bz*(uh[ijk-nf2] + uh[ijk+nf2]) +
	      by*(uh[ijk-nf1] + uh[ijk+nf1]) +
	      bx*(uh[ijk-1]   + uh[ijk+1]);
#endif
	t1  = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) * CW;
//h	if (i==nz || j==ny || k==nx)  t1=0.0;		// inefficient

       	DH[IJK-nc2-nc1-1] += t8;
        DH[IJK-nc2-nc1]   += t8+t4a;
        DH[IJK-nc2-1]     += t8+t4b;
        DH[IJK-nc1-1]     += t8+t4c;
        DH[IJK-nc2]       += t8+t4a+t4b+t2a;
        DH[IJK-nc1]       += t8+t4a+t4c+t2b;
        DH[IJK-1]         += t8+t4b+t4c+t2c;
        DH[IJK]           += t8+t4a+t4b+t4c+t2a+t2b+t2c+t1;

	k += 2;
      }
      j += 2;
    }
    i += 2;
  }
//  CheckBoundaryZero(uh);			// OK
//  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
//					Set boundary to Zero !!
  coarsegrid.SetBoundaryZero(DH);
}

//---------------------------------------------------------------------------

//void DiagLaplaceEqui3D :: Mult_Dirichlet(double *f, const double *u) const
void DiagLaplaceEqui3D :: Mult(double f[], const double u[]) const
{
 double 	t;
 unsigned int	i,j,k,n1,n2,jj;

 n1 = Dim1();    n2 = Dim2();

 for (jj=idx(0,0,0);  jj<=idx(0,ny,nx);  ++jj)  f[jj] = u[jj];
 for (i=1; i < nz; i++)
 {
   for (jj=idx(i,0,0);  jj<=idx(i,0,nx);  ++jj)  f[jj] = u[jj];
   for (j=1; j < ny; ++j)
   {
     jj = idx(i,j,1);
     f[jj-1] = u[jj-1];		// Dirichlet
     for (k=1; k < nx; ++k, ++jj)
     {
       t = bz*(u[jj-n2] + u[jj+n2]) +
	   by*(u[jj-n1] + u[jj+n1]) +
	   bx*(u[jj- 1] + u[jj +1]);

       f[jj] = diag[jj] * u[jj] + t;
     }
     f[jj] = u[jj];		// Dirichlet
   }
   for (jj=idx(i,ny,0);  jj<=idx(i,ny,nx);  jj++)  f[jj] = u[jj];
 }
 for (jj=idx(nz,0,0);  jj<=idx(nz,ny,nx);  jj++)  f[jj] = u[jj];
 return;
}

//---------------------------------------------------------------------------
//
// //void DiagLaplaceEqui3D :: Mult_Neumann(double *f, const double *u) const
// void DiagLaplaceEqui3D :: Mult(double *f, const double *u) const
// {
//  double 	t;
//  unsigned int	i,j,k,n1,n2,jj;
//
//  n1 = Dim1();    n2 = Dim2();
//
// //	#	#	#	#	#	#	#	#	#
// //	First plain
//
//  i=0;
// //   				First row in plain
//    jj=idx(i,0,0);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj+n2] +
// 	   			       by*u[jj+n1] +
// 	   			       bx*u[jj +1];
//    for (jj=idx(i,0,1);  jj<idx(i,0,nx);  jj++)
//    {
//      f[jj] = (diag[jj]+by+bz)*u[jj] + bz*u[jj+n2] +
// 	   			      by*u[jj+n1] +
// 	   			      bx*(u[jj- 1] + u[jj +1]);
//    }
//    jj = idx(i,0,nx);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj+n2]  +
// 	   			       by*u[jj+n1] +
// 	   			       bx*u[jj -1];
//
// //				Inner rows in plain
//    for (j=1; j < ny; j++)
//    {
//      jj = idx(i,j,0);
//      f[jj] = (diag[jj]+bx+bz)*u[jj] + bz*u[jj+n2] +
// 	   			      by*(u[jj-n1] + u[jj+n1]) +
// 	   			      bx*u[jj +1];
//      jj++;
//      for (k=1; k < nx; k++, jj++)
//      {
//        t = bz*u[jj+n2] +
// 	   by*(u[jj-n1] + u[jj+n1]) +
// 	   bx*(u[jj- 1] + u[jj +1]);
//
//        f[jj] = (diag[jj]+bz) * u[jj] + t;
//      }
//      f[jj] = (diag[jj]+bx+bz)*u[jj] + bz*u[jj+n2] +
// 	   			      by*(u[jj-n1] + u[jj+n1]) +
// 	   			      bx*u[jj -1];
//    }
//
// //				Last row in plain
//
//    jj=idx(i,ny,0);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj+n2] +
// 	   			       by*u[jj-n1] +
// 	   			       bx*u[jj +1];
//    for (jj=idx(i,ny,1);  jj<idx(i,ny,nx);  jj++)
//    {
//      f[jj] = (diag[jj]+by+bz)*u[jj] + bz*u[jj+n2] +
// 	   			      by*u[jj-n1] +
// 	   			      bx*(u[jj- 1] + u[jj +1]);
//    }
//    jj = idx(i,ny,nx);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj+n2] +
// 	   			       by*u[jj-n1] +
// 	   			       bx*u[jj -1];
//
// //	#	#	#	#	#	#	#	#	#
// //	Inner plains
//
//  for (i=1; i < nz; i++)
//  {
// //   				First row in plain
//    jj=idx(i,0,0);
//    f[jj] = (diag[jj]+bx+by)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			    by*u[jj+n1] +
// 	   			    bx*u[jj +1];
//    for (jj=idx(i,0,1);  jj<idx(i,0,nx);  jj++)
//    {
//      f[jj] = (diag[jj]+by)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			   by*u[jj+n1] +
// 	   			   bx*(u[jj- 1] + u[jj +1]);
//    }
//    jj = idx(i,0,nx);
//    f[jj] = (diag[jj]+bx+by)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			    by*u[jj+n1] +
// 	   			    bx*u[jj -1];
//
// //				Inner rows in plain
//    for (j=1; j < ny; j++)
//    {
//      jj = idx(i,j,0);
//      f[jj] = (diag[jj]+bx)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			   by*(u[jj-n1] + u[jj+n1]) +
// 	   			   bx*u[jj +1];
//      jj++;
//      for (k=1; k < nx; k++, jj++)
//      {
//        t = bz*(u[jj-n2] + u[jj+n2]) +
// 	   by*(u[jj-n1] + u[jj+n1]) +
// 	   bx*(u[jj- 1] + u[jj +1]);
//
//        f[jj] = diag[jj] * u[jj] + t;
//      }
//      f[jj] = (diag[jj]+bx)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			   by*(u[jj-n1] + u[jj+n1]) +
// 	   			   bx*u[jj -1];
//    }
//
// //				Last row in plain
//
//    jj=idx(i,ny,0);
//    f[jj] = (diag[jj]+bx+by)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			    by*u[jj-n1] +
// 	   			    bx*u[jj +1];
//    for (jj=idx(i,ny,1);  jj<idx(i,ny,nx);  jj++)
//    {
//      f[jj] = (diag[jj]+by)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			   by*u[jj-n1] +
// 	   			   bx*(u[jj- 1] + u[jj +1]);
//    }
//    jj = idx(i,ny,nx);
//    f[jj] = (diag[jj]+bx+by)*u[jj] + bz*(u[jj-n2] + u[jj+n2]) +
// 	   			    by*u[jj-n1] +
// 	   			    bx*u[jj -1];
//  }
//
//
// //	#	#	#	#	#	#	#	#	#
// //	Last plain
//
//  i=nz;
// //   				First row in plain
//    jj=idx(i,0,0);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj-n2] +
// 	   			       by*u[jj+n1] +
// 	   			       bx*u[jj +1];
//    for (jj=idx(i,0,1);  jj<idx(i,0,nx);  jj++)
//    {
//      f[jj] = (diag[jj]+by+bz)*u[jj] + bz*u[jj-n2] +
// 	   			      by*u[jj+n1] +
// 	   			      bx*(u[jj- 1] + u[jj +1]);
//    }
//    jj = idx(i,0,nx);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj-n2]  +
// 	   			       by*u[jj+n1] +
// 	   			       bx*u[jj -1];
//
// //				Inner rows in plain
//    for (j=1; j < ny; j++)
//    {
//      jj = idx(i,j,0);
//      f[jj] = (diag[jj]+bx+bz)*u[jj] + bz*u[jj-n2] +
// 	   			      by*(u[jj-n1] + u[jj+n1]) +
// 	   			      bx*u[jj +1];
//      jj++;
//      for (k=1; k < nx; k++, jj++)
//      {
//        t = bz*u[jj-n2] +
// 	   by*(u[jj-n1] + u[jj+n1]) +
// 	   bx*(u[jj- 1] + u[jj +1]);
//
//        f[jj] = (diag[jj]+bz) * u[jj] + t;
//      }
//      f[jj] = (diag[jj]+bx+bz)*u[jj] + bz*u[jj-n2] +
// 	   			      by*(u[jj-n1] + u[jj+n1]) +
// 	   			      bx*u[jj -1];
//    }
//
// //				Last row in plain
//
//    jj=idx(i,ny,0);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj-n2] +
// 	   			       by*u[jj-n1] +
// 	   			       bx*u[jj +1];
//    for (jj=idx(i,ny,1);  jj<idx(i,ny,nx);  jj++)
//    {
//      f[jj] = (diag[jj]+by+bz)*u[jj] + bz*u[jj-n2] +
// 	   			      by*u[jj-n1] +
// 	   			      bx*(u[jj- 1] + u[jj +1]);
//    }
//    jj = idx(i,ny,nx);
//    f[jj] = (diag[jj]+bx+by+bz)*u[jj] + bz*u[jj-n2] +
// 	   			       by*u[jj-n1] +
// 	   			       bx*u[jj -1];
//
//  return;
// }

//---------------------------------------------------------------------------

void DiagLaplaceEqui3D ::
Defect(const double f[], const double u[], double def[]) const
{
 double 	t;
 unsigned int	i,j,k,n1,n2,jj;

 SetBoundaryZero(def);			// homogeneous Dirichlet B.C.

 n1 = nx+1;    n2 = (nx+1)*(ny+1);

 for (i=1; i < nz; ++i)
 {
   for (j=1; j < ny; ++j)
   {
     jj = idx(n2,n1,i,j,1);
     for (k=1; k < nx; ++k, ++jj)
     {
       t = bz*(u[jj-n2] + u[jj+n2]) +
	   by*(u[jj-n1] + u[jj+n1]) +
	   bx*(u[jj- 1] + u[jj +1]);

       def[jj] = f[jj] - diag[jj] * u[jj] - t;
     }
   }
 }
 return;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3D :: Jacobi(double u[], const double f[], double def[],
  			const double omega, const int iter) const
{
 int n = VectorLength();

 Defect(f,u,def);
 VDdiv(n,def,diag);
 VDXpaY(n,u,omega,def);
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3D ::
GaussSeidelForward(double u[], const double f[], const int iter) const
{
  double 	t;
  unsigned int	i,j,k,n1,n2,jj;

  n1 = nx+1;
  n2 = (nx+1)*(ny+1);

  for (i=1; i < nz; ++i)
  {
    for (j=1; j < ny; ++j)
    {
      jj = idx(n2,n1,i,j,1);
      for (k=1; k < nx; ++k, ++jj)
      {
	t = bz*(u[jj-n2] + u[jj+n2]) +
	    by*(u[jj-n1] + u[jj+n1]) +
	    bx*(u[jj-1]  + u[jj+1] );

        u[jj]= (f[jj] - t) / diag[jj];
      }
    }
  }
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
}
//---------------------------------------------------------------------------

void DiagLaplaceEqui3D ::
GaussSeidelBackward(double u[], const double f[], const int iter) const
{
  double 	t;
  unsigned int	i,j,k,n1,n2,jj;

  n1 = nx+1;
  n2 = (nx+1)*(ny+1);

  for (i=nz-1; i > 0; --i)
  {
    for (j=ny-1; j > 0; --j)
    {
      jj = idx(n2,n1,i,j,nx-1);
      for (k=nx-1; k > 0; --k, --jj)
      {
	t = bz*(u[jj-n2] + u[jj+n2]) +
	    by*(u[jj-n1] + u[jj+n1]) +
	    bx*(u[jj-1]  + u[jj+1] );

        u[jj]= (f[jj] - t) / diag[jj];
      }
    }
  }
}

//---------------------------------------------------------------------------

double DiagLaplaceEqui3D :: DefectNorm(const double u[], const double f[]) const
{
  double 	t,s,r;
  unsigned int	i,j,k,n1,n2,jj;

  n1 = nx+1;
  n2 = (nx+1)*(ny+1);

  r = 0;
  for (i=1; i < nz; ++i)
  {
    for (j=1; j < ny; ++j)
    {
      jj = idx(n2,n1,i,j,1);
      for (k=1; k < nx; ++k, ++jj)
      {
	t = bz*(u[jj-n2] + u[jj+n2]) +
	    by*(u[jj-n1] + u[jj+n1]) +
	    bx*(u[jj-1]  + u[jj+1] );

        s = f[jj] - diag[jj] * u[jj] - t;
        r += s*s;
      }
    }
  }

  return(sqrt(r));
 }

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., Laplace + c
//	rho*I = sqrt(2)*Laplace_y + c/2 *I

void DiagLaplaceEqui3D :: ADIx(double u[], const double f[]) const
{
  cout << "#########  DiagLaplaceEqui3D :: ADIx   is not implemented  ###########" << endl;
//   const double CSQ2 = 1.0-sqrt(2.0);
//   const    int n1 = Dim1();
//   const    int nfx = GetIntX(),  nfy = GetIntY();
// //old  const double rho = sqrt(8.0)*nfy*nfy;
//         double tm;
//            int i,j,jj;
//
//   for (i=0; i <= nfx; i++)   u[i] = 0.0;
//   for (i=0; i <= nfx; i++)  tt[i] = 0.0;
//   for (j=1; j < nfy; j++)
//   {
//    jj = idx(j,0);
//    tm = u[jj];			//	right hand side
//    u[jj] = 0.0;				// Dirichlet
//    for (i=1; i < nfx; i++)
//      {					// f + (rho-D_y)tm ...
//        tm   = u[jj+i];
// //old       u[jj+i] = f[jj+i] + (rho+2*by)*tm - by*(u[jj+n1+i]+tt[i]);
// //       u[jj+i] = f[jj+i] + 2*CSQ2*by*tm - by*(u[jj+n1+i]+tt[i]);
//        u[jj+i] = f[jj+i] + by*(2*CSQ2*tm - u[jj+n1+i]-tt[i]);
//        tt[i] = tm;
//      }
//    u[jj+nfx] = 0.0;
//
//    SubstFactorSymmTriDiag(nfx+1, diagx, offx, u+jj);
//   }
//   for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;
//
  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C.,  Laplace + c
//	rho*I = sqrt(2)*Laplace_x + c/2 *I

void DiagLaplaceEqui3D :: ADIy(double u[], const double f[]) const
{
  cout << "#########  DiagLaplaceEqui3D :: ADIy   is not implemented  ###########" << endl;
//   const double CSQ2 = 1.0-sqrt(2.0);
//   const    int n1 = Dim1();
//   const    int nfx = GetIntX(),  nfy = GetIntY();
// //old  const double rho = sqrt(8.0)*nfx*nfx;
//         double tm,tb;
//            int i,j,jj;
//
//   for (i=0; i <= nfx; i++)  u[i] = 0.0;
//   for (j=1; j < nfy; j++)
//   {
//     jj = idx(j,0);		//	right hand side
//     tm = u[jj];
//     u[jj] = 0.0;			// Dirichlet
//     for (i=1; i < nfx; i++)
//      {					// f + (rho-D_x)tm ...
//        tb = tm;
//        tm = u[jj+i];
// //old       u[jj+i] = f[jj+i] + (rho+2*bx)*tm - bx*(u[jj+i+1]+tb);
// //       u[jj+i] = f[jj+i] + 2*CSQ2*bx*tm - bx*(u[jj+i+1]+tb);
//        u[jj+i] = f[jj+i] + bx*(2*CSQ2*tm - u[jj+i+1]-tb);
//      }
//     u[jj+nfx] = 0.0;			// Dirichlet
//   }
//   for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;
//
//   SubstFactorSymmTriDiag_MultT(nfy+1, diagy, offy, u, n1);
//
  return;
}

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void DiagLaplaceEqui3D ::
ADIForward(double *u, const double *f, double *aux, const int iter) const
{
#ifdef DEBUG
  cout << "#########  u_in  ###########" << endl;
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
  cout << "#########  f  ###########" << endl;
  PrintVector(VectorLength(),f);
  CheckBoundaryZero(f);
#endif

  ADIx(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
#endif

  ADIy(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
#endif
  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void DiagLaplaceEqui3D ::
ADIBackward(double *u, const double *f, double *aux, const int iter) const
{
 ADIy(u,f);
 ADIx(u,f);
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
  return;
}

//---------------------------------------------------------------------------
void DiagLaplaceEqui3D :: FactorADIMatrices()
{
//  const double CSQ2 = 1.0-sqrt(2.0);
//  double       rho,c2,diagsum;
//  int          i,j, nfx=GetIntX(), nfy=GetIntY();
//
//  if (diagx!=0) delete [] diagx;
//  diagx  = new double [nfx+1];
//  if (offx!=0) delete [] offx;
//  offx   = new double [nfx];
//
// //	rho_x*I = sqrt(2)*Laplace_x + c/2 *I
// //old rho = sqrt(8.0)*nfy*nfy;
// 				//	tridiagonal symmetric matrix
//  diagx[0] = diagx[nfx] = 1.0;	// Dirichlet
//  for (i=1; i<nfx; i++)		// diagonal entries
//   {				// diag = rho*I + D_x
//     diagsum=0.0;
//     for (j=1; j<nfy; j++)  diagsum += diag[idx(j,i)];
//       diagx[i] = diagsum/(nfy-1) + 2*CSQ2*by;
// //old    diagx[i] = rho-2*bx;
//   }
//
//  offx[0] = offx[nfx-1] = 0.0;	// Dirichlet
//  for (i=1; i<nfx-1; i++)	// off-digonal entries
//      offx[i] =  bx;
//
//  FactorSymmTriDiag(nfx+1, diagx, offx);
//
//  if (diagy!=0) delete [] diagy;
//  diagy  = new double [nfy+1];
//  if (offy!=0) delete [] offy;
//  offy   = new double [nfy];
//
// //	rho_y*I = sqrt(2)*Laplace_y + c/2 *I
// //old rho = sqrt(8.0)*nfx*nfx;
// 				//	tridiagonal symmetric matrix
//  diagy[0] = diagy[nfy] = 1.0;			// Dirichlet
//  for (i=1; i<nfy; i++)
//   {
//     diagsum=0.0;
//     for (j=1; j<nfx; j++)  diagsum += diag[idx(i,j)];
//     diagy[i] = diagsum/(nfx-1)+2*CSQ2*bx;		// diagonal
// //old    diagy[i] = rho-2*by;				// diagonal
//   }
//
//  offy[0] = offy[nfy-1] = 0.0;			// Dirichlet
//  for (i=1; i<nfy-1; i++)
//      offy[i] =  by;					// off-digonal
//
//  FactorSymmTriDiag(nfy+1, diagy, offy);

 return;
}


//	###################################################################
//			LaplaceNonEqui3D
//	###################################################################
//	Laplace 3D, non-equidistant tensor product grid

//---------------------------------------------------------------------------

LaplaceNonEqui3D ::
LaplaceNonEqui3D(const int nnx, const double x[],
                 const int nny, const double y[],
	         const int nnz, const double z[], const bool coarse)
 : NonEqui3D(nnx,x,nny,y,nnz,z),
   diag(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
//   diag = 0;
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

LaplaceNonEqui3D ::
LaplaceNonEqui3D(const int nnx, const double *x, const bool coarse)
 : NonEqui3D(nnx,x), diag(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
//   if (x==0)
//    {
//      diag = bxr = bxl = byr = byl = bzr = bzl = 0;
//    }
//   else
//    {
//      diag = 0;
     InitOperator(coarse);
//    }
}
//---------------------------------------------------------------------------
LaplaceNonEqui3D ::
LaplaceNonEqui3D(const int intx, const int inty, const int intz,
                 const double xh[], const double yh[], const double zh[],
	         const bool coarse)
 : NonEqui3D(intx, inty, intz, xh, yh, zh),
   diag(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
//   diag = 0;
  InitOperator(coarse);
}
//---------------------------------------------------------------------------

 LaplaceNonEqui3D :: ~LaplaceNonEqui3D()
 {
  DeleteGridVector(diag);
  delete [] bxl; delete [] bxr;
  delete [] byl; delete [] byr;
  delete [] bzl; delete [] bzr;
 }

//---------------------------------------------------------------------------

LaplaceNonEqui3D :: LaplaceNonEqui3D(const LaplaceNonEqui3D &orig)
: NonEqui3D(orig),
   diag(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
 NewGridVector(diag);
 VDcopy( VectorLength(), diag, 1, orig.diag, 1 );

 bxl = new double [GetIntX()+1];
 bxr = new double [GetIntX()+1];
 VDcopy( GetIntX()+1, bxl, 1, orig.bxl, 1 );
 VDcopy( GetIntX()+1, bxr, 1, orig.bxr, 1 );
 byl = new double [GetIntY()+1];
 byr = new double [GetIntY()+1];
 VDcopy( GetIntY()+1, byl, 1, orig.byl, 1 );
 VDcopy( GetIntY()+1, byr, 1, orig.byr, 1 );
 bzl = new double [GetIntZ()+1];
 bzr = new double [GetIntZ()+1];
 VDcopy( GetIntZ()+1, bzl, 1, orig.bzl, 1 );
 VDcopy( GetIntZ()+1, bzr, 1, orig.bzr, 1 );
}

//---------------------------------------------------------------------------
LaplaceNonEqui3D& LaplaceNonEqui3D :: operator=(const LaplaceNonEqui3D & orig)
{
 if ( this != &orig )
 {
   NonEqui3D::operator=(orig);

  if (diag != 0)
   {
     DeleteGridVector(diag);
     delete [] bxl; delete [] bxr;
     delete [] byl; delete [] byr;
     delete [] bzl; delete [] bzr;
   }

  if (orig.diag != 0)
   {
     NewGridVector(diag);
     VDcopy( VectorLength(), diag, 1, orig.diag, 1 );

     bxl = new double [GetIntX()+1];
     bxr = new double [GetIntX()+1];
     VDcopy( GetIntX()+1, bxl, 1, orig.bxl, 1 );
     VDcopy( GetIntX()+1, bxr, 1, orig.bxr, 1 );
     byl = new double [GetIntY()+1];
     byr = new double [GetIntY()+1];
     VDcopy( GetIntY()+1, byl, 1, orig.byl, 1 );
     VDcopy( GetIntY(), byr, 1, orig.byr, 1 );
     bzl = new double [GetIntZ()+1];
     bzr = new double [GetIntZ()+1];
     VDcopy( GetIntZ()+1, bzl, 1, orig.bzl, 1 );
     VDcopy( GetIntZ()+1, bzr, 1, orig.bzr, 1 );
   }
 }
 return *this;
}

//---------------------------------------------------------------------------
void LaplaceNonEqui3D :: Refine(LaplaceNonEqui3D & fine) const
{
  LaplaceNonEqui3D tmpgrid;
  NonEqui3D::Refine(tmpgrid);

  tmpgrid.coarsegrid_ = this;
  tmpgrid.diag = 0;
  tmpgrid.InitOperator(false);
  fine = tmpgrid;

//  fine.DeriveMatrix()  must not be executed since this is already a fine grid ;
}

//---------------------------------------------------------------------------
void LaplaceNonEqui3D ::OffDiag1D
	(const int n, const double h[], double bl[], double br[])
{
 int i;

 br[0] = 0.0;						// Dirichlet-RB
// br[0] = -2.0/(h[0]+h[1])/h[1] = -2.0/h[1]/h[1];	// Neumann-RB
 bl[0] = 0.0;
 for  (i=1; i < n; i++)
  {
//				[Grossmann/Roos] ==> non-symmetric matrix
  br[i] = -2.0/(h[i]+h[i+1])/h[i+1];
  bl[i] = -2.0/(h[i]+h[i+1])/h[i];
//				Bloedsinn, aber symmetrisch
//    br[i] = -1.0/h[i+1]/h[i+1];
//    bl[i] = -1.0/h[i]/h[i];
//    cout <<  i << "::: "<< bl[i] << " "<< br[i]<< "          " << h[i] << " " << h[i+1] << endl;
//				Bloedsinn, aber symmetrisch
//   br[i] = -2.0/(h[i+1]+h[i+2])/h[i+1];
//   bl[i] = -2.0/(h[i]+h[i+1])/h[i];
  }
 br[n] = 0.0;
 bl[n] = 0.0;						// Dirichlet-RB
// bl[n] = -2.0/(h[n]+h[n+1])/h[n] = -2.0/h[n]/h[n];	// Neumann-RB
 return;
}
//---------------------------------------------------------------------------

void LaplaceNonEqui3D ::InitOperator(const bool coarse, const double *cc)
{
  int n   = VectorLength();
  int nhx = GetIntX(),  nhy = GetIntY(),  nhz = GetIntZ();
  int i,j,k,ijk;
  double tk,tj;

  if (diag==0)
   {
     NewGridVector(diag);
     bxl = new double [nhx+1];
     bxr = new double [nhx+1];
     byl = new double [nhy+1];
     byr = new double [nhy+1];
     bzl = new double [nhz+1];
     bzr = new double [nhz+1];
   }

  OffDiag1D(nhx, hhx, bxl, bxr);
  OffDiag1D(nhy, hhy, byl, byr);
  OffDiag1D(nhz, hhz, bzl, bzr);


  VDcopyC( n, diag, 0.0 );
  for (i=0, ijk = 0; i<=nhz; i++)
   {
    tk = bzl[i]+bzr[i];
    for (j=0; j<=nhy; j++)
     {
      tj = byl[j]+byr[j];
      for (k=0; k<=nhx; k++, ijk++)
       {
         diag[ijk] = -bxl[k]-bxr[k] - tj - tk;
       }
     }
   }

  if  (cc !=0 ) 	//			Add Helmholtz  term
  	VDadd( n, diag, cc );

  SetBoundaryConst(diag, 1.0);		// Whole boundary is Dirichlet

  if (coarse)
    {			//	diag is used in matrix calculation
      DeriveMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Original Coarse Matrix" << endl;
      PrintMatrix(n, mat);
      bool bb=CheckMatrixSymmetry(n, mat);
      cout << "############################" << endl;
#endif
      InvertMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Inverted Coarse Matrix" << endl;
      PrintMatrix(n, mat);
#endif
    }
}

//---------------------------------------------------------------------------
//	Derive matrix for the 7 point stencil
//		and Dirichlet boundary conditions

void LaplaceNonEqui3D :: DeriveMatrix()
{
 int i,j,k,n1,n2,ijk,id;
 int nn = VectorLength();
 unsigned long int nn2 = nn*nn;

 n1 = nx+1;  n2 = (nx+1)*(ny+1);

//	Allocate memory for fully occupied matrix if called the first time
//			(Assumes no changes in dimensions!!)
 if (mat==0)   mat = new double[nn2];
 VDcopyC(nn2, mat, 0.0);	// set matrix to Zero
	// Initialize diagonal with 1  <== Dirichlet boundary
 for ( i=0; i<nn; i++ )  mat[i*(nn+1)] = 1.0;

 for ( i=1; i<nz; i++ )
  {
   for ( j=1; j<ny; j++ )
    {
     for ( k=1; k<nx; k++ )
      {
        ijk = idx(n2,n1,i,j, k);
	id  = ijk*nn + ijk;
	mat[id] = diag[ijk];
	if ( k != 1    ) mat[id-1 ] = bxl[k];
	if ( k != nx-1 ) mat[id+1 ] = bxr[k];
	if ( j != 1    ) mat[id-n1] = byl[j];
 	if ( j != ny-1 ) mat[id+n1] = byr[j];
	if ( i != 1    ) mat[id-n2] = bzl[i];
 	if ( i != nz-1 ) mat[id+n2] = bzr[i];
     }
    }
  }
#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix" << endl;
  PrintMatrix(nn,mat);
#endif
}

//---------------------------------------------------------------------------

void LaplaceNonEqui3D :: DirectSolve(double *u, const double *f) const
{
 if ( nx != 2 ||  ny != 2 || nz != 2)
  {
   NonEqui3D::DirectSolve(u,f);
  }
 else
  {
   u[13] = f[13] / diag[13];
  }
}

//---------------------------------------------------------------------------

void LaplaceNonEqui3D :: DefRes(const double *fh, const double *uh, double *DH) const
{
  const double  CW = 1./8.;			//  normal weight

  double 	t,t1,t2yx,t2zx,t2zy,t4x,t4y,t4z,t8;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const LaplaceNonEqui3D   &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = Dim1();              nf2 = Dim2();
  nc1 = coarsegrid.Dim1();   nc2 = coarsegrid.Dim2();

  VDcopyC(coarsegrid.VectorLength(), DH, 0.0);

  i = 2;
  for (ii=1; ii <= ncz; ii++)
  {
   j = 2;
   for (jj=1; jj <= ncy; jj++)
    {
      k = 2;
      IJK = idx(nc2,nc1,ii,jj, 1);
      for (kk=1; kk <= ncx; kk++, IJK++)
      {
//			Compiler besser untersttzen!!
//
	ijk = idx(nf2,nf1,i-1,j-1,k-1);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	      byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	      bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t8  = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

	ijk = idx(nf2,nf1,i-1,j-1,k);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	      byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	      bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
        t4x = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (k==nx)  t4x=0.0;		// inefficient

	ijk = idx(nf2,nf1,i-1,j,k-1);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	      byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	      bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t4y = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (j==ny)  t4y=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j-1,k-1);
        t   = bzl[i  ]*uh[ijk-nf2] + bzr[i  ]*uh[ijk+nf2] +
	      byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	      bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t4z = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz)  t4z=0.0;		// inefficient

	ijk = idx(nf2,nf1,i-1,j,k);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	      byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	      bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
        t2yx = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (k==nx || j==ny)  t2yx=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j-1,k);
        t   = bzl[i  ]*uh[ijk-nf2] + bzr[i  ]*uh[ijk+nf2] +
	      byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	      bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
        t2zx = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz || k==nx)  t2zx=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j,k-1);
        t   = bzl[i  ]*uh[ijk-nf2] + bzr[i  ]*uh[ijk+nf2] +
	      byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	      bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t2zy = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz || j==ny)  t2zy=0.0;		// inefficient

 	ijk = idx(nf2,nf1,i,j,k);
        t   = bzl[i]*uh[ijk-nf2] + bzr[i]*uh[ijk+nf2] +
	      byl[j]*uh[ijk-nf1] + byr[j]*uh[ijk+nf1] +
	      bxl[k]*uh[ijk-1]   + bxr[k]*uh[ijk+1];
        t1  = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz || j==ny || k==nx)  t1=0.0;		// inefficient

       	DH[IJK-nc2-nc1-1] += wxl(k-1)*wyl(j-1)*wzl(i-1)*t8;
        DH[IJK-nc2-nc1]   += wxr(k-1)*wyl(j-1)*wzl(i-1)*t8
			    +         wyl(j-1)*wzl(i-1)*t4x;
        DH[IJK-nc2-1]     += wxl(k-1)*wyr(j-1)*wzl(i-1)*t8
			    +wxl(k-1)         *wzl(i-1)*t4y;
        DH[IJK-nc1-1]     += wxl(k-1)*wyl(j-1)*wzr(i-1)*t8
	                    +wxl(k-1)*wyl(j-1)         *t4z;
        DH[IJK-nc2]       += wxr(k-1)*wyr(j-1)*wzl(i-1)*t8
	                    +         wyr(j-1)*wzl(i-1)*t4x
			    +wxr(k-1)         *wzl(i-1)*t4y
			    +                  wzl(i-1)*t2yx;
        DH[IJK-nc1]       += wxr(k-1)*wyl(j-1)*wzr(i-1)*t8
	                    +         wyl(j-1)*wzr(i-1)*t4x
			    +wxr(k-1)*wyl(j-1)         *t4z
			    +         wyl(j-1)         *t2zx;
        DH[IJK-1]         += wxl(k-1)*wyr(j-1)*wzr(i-1)*t8
	                    +wxl(k-1)         *wzr(i-1)*t4y
			    +wxl(k-1)*wyr(j-1)         *t4z
			    +wxl(k-1)                  *t2zy;
        DH[IJK]           += wxr(k-1)*wyr(j-1)*wzr(i-1)*t8
	                    +         wyr(j-1)*wzr(i-1)*t4x
			    +wxr(k-1)         *wzr(i-1)*t4y
			    +wxr(k-1)*wyr(j-1)         *t4z
			    +                  wzr(i-1)*t2yx
			    +         wyr(j-1)         *t2zx
			    +wxr(k-1)                  *t2zy
			    +t1;

	k += 2;
      }
      j += 2;
    }
    i += 2;
  }
//  CheckBoundaryZero(uh);			// OK
//  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
//					Set boundary to Zero !!
  coarsegrid.SetBoundaryZero(DH);
}

//---------------------------------------------------------------------------
//		What happens when we program the method in vector style !??!

void LaplaceNonEqui3D :: Mult(double *f, const double *u) const
{
 double 	t,tjl,tjr,til,tir;
 unsigned int	i,j,k,n1,n2,jj;

 n1 = Dim1();    n2 = Dim2();

 for (jj=idx(0,0,0);  jj<=idx(0,ny,nx);  jj++)  f[jj] = u[jj];
 for (i=1; i < nz; i++)
 {
   for (jj=idx(i,0,0);  jj<=idx(i,0,nx);  jj++)  f[jj] = u[jj];
   til = bzl[i];  tir = bzr[i];
   for (j=1; j < ny; j++)
   {
     tjl = byl[j];  tjr = byr[j];
     jj  = idx(i,j,1);
     f[jj-1] = u[jj-1];		// Dirichlet
     for (k=1; k < nx; k++, jj++)
     {
       t = til   * u[jj-n2] + tir   * u[jj+n2] +
	   tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	   bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

       f[jj] = diag[jj] * u[jj] + t;
     }
     f[jj] = u[jj];		// Dirichlet
   }
   for (jj=idx(i,ny,0);  jj<=idx(i,ny,nx);  jj++)  f[jj] = u[jj];
 }
 for (jj=idx(nz,0,0);  jj<=idx(nz,ny,nx);  jj++)  f[jj] = u[jj];
}

//---------------------------------------------------------------------------
//		What happens when we program the method in vector style !??!

void LaplaceNonEqui3D ::
Defect(const double *f, const double *u, double *def) const
{
 double 	t,tjl,tjr,til,tir;
 unsigned int	i,j,k,n1,n2,jj;

 SetBoundaryZero(def);			// homogeneous Dirichlet B.C.

 n1 = Dim1();    n2 = Dim2();

 for (i=1; i < nz; i++)
 {
   til = bzl[i];  tir = bzr[i];
   for (j=1; j < ny; j++)
   {
     tjl = byl[j];  tjr = byr[j];
     jj  = idx(n2,n1,i,j,1);
     for (k=1; k < nx; k++, jj++)
     {
       t = til   * u[jj-n2] + tir   * u[jj+n2] +
	   tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	   bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

       def[jj] = f[jj] - diag[jj] * u[jj] - t;
     }
   }
 }
}

//---------------------------------------------------------------------------

void LaplaceNonEqui3D :: Jacobi(double *u, const double *f, double *def,
  			const double omega, const int iter) const
{
 int n = VectorLength();

 Defect(f,u,def);
 VDdiv(n,def,diag);
 VDXpaY(n,u,omega,def);
}

//---------------------------------------------------------------------------

void LaplaceNonEqui3D ::
GaussSeidelForward(double *u, const double *f, const int iter) const
{
  double 	t,til,tir,tjl,tjr;
  unsigned int	i,j,k,jj,n1,n2;

  n1 = Dim1(); n2 = Dim2();

  for (i=1; i < nz; i++)
  {
    til = bzl[i];  tir = bzr[i];
    for (j=1; j < ny; j++)
    {
      tjl = byl[j];  tjr = byr[j];
      jj = idx(n2,n1,i,j,1);
      for (k=1; k < nx; k++, jj++)
      {
	t = til   * u[jj-n2] + tir   * u[jj+n2] +
	    tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	    bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

        u[jj]= (f[jj] - t) / diag[jj];
      }
    }
  }
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
}
//---------------------------------------------------------------------------

void LaplaceNonEqui3D ::
GaussSeidelBackward(double *u, const double *f, const int iter) const
{
  double 	t,til,tir,tjl,tjr;
  unsigned int	i,j,k,n1,n2,jj;

  n1 = Dim1(); n2 = Dim2();

  for (i=nz-1; i > 0; i--)
  {
    til = bzl[i];  tir = bzr[i];
    for (j=ny-1; j > 0; j--)
    {
      tjl = byl[j];  tjr = byr[j];
      jj = idx(n2,n1,i,j,nx-1);
      for (k=nx-1; k > 0; k--, jj--)
      {
	t = til   * u[jj-n2] + tir   * u[jj+n2] +
	    tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	    bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

        u[jj]= (f[jj] - t) / diag[jj];
      }
    }
  }
}

//---------------------------------------------------------------------------

double LaplaceNonEqui3D :: DefectNorm(const double *u, const double *f) const
{
  double 	t,s,r,til,tir,tjl,tjr;
  unsigned int	i,j,k,n1,n2,jj;

  n1 = Dim1(); n2 = Dim2();

  r = 0.0;
  for (i=1; i < nz; i++)
  {
    til = bzl[i];  tir = bzr[i];
    for (j=1; j < ny; j++)
    {
      tjl = byl[j];  tjr = byr[j];
      jj = idx(n2,n1,i,j,1);
      for (k=1; k < nx; k++, jj++)
      {
	t = til   * u[jj-n2] + tir   * u[jj+n2] +
	    tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	    bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

        s = f[jj] - diag[jj] * u[jj] - t;
        r += s*s;
      }
    }
  }

//   cout << "-------------------"<<endl;
//   for (i=1; i<nz; i++)
//      cout << i << ":: " << bzl[i] << " " << bzr[i]
//       << "            " << hhz[i] << " " << hhz[i+1]
//       << endl;

  return(sqrt(r));
 }

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., Laplace + c
//	rho*I = sqrt(2)*Laplace_y + c/2 *I

void LaplaceNonEqui3D :: ADIx(double u[], const double f[]) const
{
//   const double CSQ2 = 1.0-sqrt(2.0);
//   const    int n1 = Dim1();
//   const    int nfx = GetIntX(),  nfy = GetIntY();
// //old  const double rho = sqrt(8.0)*nfy*nfy;
//         double tm;
//            int i,j,jj;
//
//   for (i=0; i <= nfx; i++)   u[i] = 0.0;
//   for (i=0; i <= nfx; i++)  tt[i] = 0.0;
//   for (j=1; j < nfy; j++)
//   {
//    jj = idx(j,0);
//    tm = u[jj];			//	right hand side
//    u[jj] = 0.0;				// Dirichlet
//    for (i=1; i < nfx; i++)
//      {					// f + (rho-D_y)tm ...
//        tm   = u[jj+i];
// //old       u[jj+i] = f[jj+i] + (rho+2*by)*tm - by*(u[jj+n1+i]+tt[i]);
// //       u[jj+i] = f[jj+i] + 2*CSQ2*by*tm - by*(u[jj+n1+i]+tt[i]);
//        u[jj+i] = f[jj+i] + by*(2*CSQ2*tm - u[jj+n1+i]-tt[i]);
//        tt[i] = tm;
//      }
//    u[jj+nfx] = 0.0;
//
//    SubstFactorSymmTriDiag(nfx+1, diagx, offx, u+jj);
//   }
//   for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;
//
  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C.,  Laplace + c
//	rho*I = sqrt(2)*Laplace_x + c/2 *I

void LaplaceNonEqui3D :: ADIy(double u[], const double f[]) const
{
//   const double CSQ2 = 1.0-sqrt(2.0);
//   const    int n1 = Dim1();
//   const    int nfx = GetIntX(),  nfy = GetIntY();
// //old  const double rho = sqrt(8.0)*nfx*nfx;
//         double tm,tb;
//            int i,j,jj;
//
//   for (i=0; i <= nfx; i++)  u[i] = 0.0;
//   for (j=1; j < nfy; j++)
//   {
//     jj = idx(j,0);		//	right hand side
//     tm = u[jj];
//     u[jj] = 0.0;			// Dirichlet
//     for (i=1; i < nfx; i++)
//      {					// f + (rho-D_x)tm ...
//        tb = tm;
//        tm = u[jj+i];
// //old       u[jj+i] = f[jj+i] + (rho+2*bx)*tm - bx*(u[jj+i+1]+tb);
// //       u[jj+i] = f[jj+i] + 2*CSQ2*bx*tm - bx*(u[jj+i+1]+tb);
//        u[jj+i] = f[jj+i] + bx*(2*CSQ2*tm - u[jj+i+1]-tb);
//      }
//     u[jj+nfx] = 0.0;			// Dirichlet
//   }
//   for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;
//
//   SubstFactorSymmTriDiag_MultT(nfy+1, diagy, offy, u, n1);
//
  return;
}

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void LaplaceNonEqui3D ::
ADIForward(double *u, const double *f, double *aux, const int iter) const
{
#ifdef DEBUG
  cout << "#########  u_in  ###########" << endl;
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
  cout << "#########  f  ###########" << endl;
  PrintVector(VectorLength(),f);
  CheckBoundaryZero(f);
#endif

  ADIx(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
#endif

  ADIy(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
#endif
  return;
}

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void LaplaceNonEqui3D ::
ADIBackward(double *u, const double *f, double *aux, const int iter) const
{
 ADIy(u,f);
 ADIx(u,f);
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
  return;
}

//---------------------------------------------------------------------------

//	###################################################################
//			DiagLaplaceEqui3DNeumann
//	###################################################################
//	Laplace 3D, equidistant tensor product grid, homogeneoeus Neumann BC

DiagLaplaceEqui3DNeumann ::
DiagLaplaceEqui3DNeumann(const int mx, const int my, const int mz, const bool coarse)
 : DiagLaplaceEqui3D(mx,my,mz,coarse)
{
  DeleteGridVector(diag);
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui3DNeumann :: DiagLaplaceEqui3DNeumann(const int mx, const bool coarse)
 : DiagLaplaceEqui3D(mx,mx,mx,coarse)
{
  DeleteGridVector(diag);
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui3DNeumann :: ~DiagLaplaceEqui3DNeumann()
{
 ;
}
//---------------------------------------------------------------------------

DiagLaplaceEqui3DNeumann :: DiagLaplaceEqui3DNeumann(const DiagLaplaceEqui3DNeumann &orig)
: DiagLaplaceEqui3D(orig)
{
 ;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3DNeumann :: Refine(DiagLaplaceEqui3DNeumann & fine) const
{
//						not a coarse grid
  fine = DiagLaplaceEqui3DNeumann(2*nx,2*ny,2*nz,false);
  fine.coarsegrid_ = this;
//  fine.DeriveMatrix()  must not be executed since this is already a fine grid ;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3DNeumann :: Mult(double *f, const double *u) const
{
 double cx,cy,cz,t;
 int i,j,k,n1,n2,ijk;

 n1 = nx+1;  n2 = (nx+1)*(ny+1);

 for ( i=0; i<=nz; i++ )
  {
   for ( j=0; j<=ny; j++ )
    {
     for ( k=0; k<=nx; k++ )
      {
	cx = bx; cy = by; cz = bz;
	if ( i==0 || i == nz )  { cx = bx/2.; cy = by/2.; }
	if ( j==0 || j == ny )  { cx = bx/2.; cz = bz/2.; }
	if ( k==0 || k == nx )  { cy = by/2.; cz = bz/2.; }

	ijk = idx(i,j,k);
	t = diag[ijk]*u[ijk];
	if ( k != 0  ) t += cx*u[ijk-1];
	if ( k != nx ) t += cx*u[ijk+1];
	if ( j != 0  ) t += cy*u[ijk-n1];
 	if ( j != ny ) t += cy*u[ijk+n1];
	if ( i != 0  ) t += cz*u[ijk-n2];
 	if ( i != nz ) t += cz*u[ijk+n2];
	f[ijk] = t;
//	if (i+j+k==0) cout << diag[ijk] << "  " << cx << "  " << cy << "  " << cz << endl;
//	if (i+j+k==0) cout << u[ijk] << "  " << u[ijk+1] << "  " << u[ijk+n1] << "  " << u[ijk+n2] << endl;
     }
    }
  }
}
//---------------------------------------------------------------------------

//void DiagLaplaceEqui3DNeumann :: Mult(double *f, const double *u) const
void DiagLaplaceEqui3DNeumann :: Mult_long(double *f, const double *u) const
{
 double 	t;
 unsigned int	i,j,k,n1,n2,jj;

 n1 = Dim1();    n2 = Dim2();

//	#	#	#	#	#	#	#	#	#
//	First plain

 i=0;
//   				First row in plain
   jj=idx(i,0,0);
   f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2. +
	   			by*u[jj+n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(i,0,1);  jj<idx(i,0,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2. +
	   			by*u[jj+n1]/2. +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(i,0,nx);
   f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2.  +
	   			by*u[jj+n1]/2. +
	   			bx*u[jj -1]/2.;

//				Inner rows in plain
   for (j=1; j < ny; j++)
   {
     jj = idx(i,j,0);
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2. +
	   			by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj +1]/2.;
     jj++;
     for (k=1; k < nx; k++, jj++)
     {
       t = bz*u[jj+n2] +
	   by*(u[jj-n1] + u[jj+n1])/2. +
	   bx*(u[jj- 1] + u[jj +1])/2.;

       f[jj] = diag[jj] * u[jj] + t;
     }
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2. +
	   			by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj -1]/2.;
   }

//				Last row in plain

   jj=idx(i,ny,0);
   f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2. +
	   			by*u[jj-n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(i,ny,1);  jj<idx(i,ny,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2. +
	   			by*u[jj-n1]/2. +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(i,ny,nx);
   f[jj] = diag[jj]*u[jj] + 	bz*u[jj+n2]/2. +
	   			by*u[jj-n1]/2. +
	   			bx*u[jj -1]/2.;

//	#	#	#	#	#	#	#	#	#
//	Inner plains

 for (i=1; i < nz; i++)
 {
//   				First row in plain
   jj=idx(i,0,0);
   f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2])/2. +
	   			by*u[jj+n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(i,0,1);  jj<idx(i,0,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2])/2. +
	   			by*u[jj+n1] +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(i,0,nx);
   f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2])/2. +
	   			by*u[jj+n1]/2. +
	   			bx*u[jj -1]/2.;

//				Inner rows in plain
   for (j=1; j < ny; j++)
   {
     jj = idx(i,j,0);
     f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2])/2. +
	   			by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj +1];
     jj++;
     for (k=1; k < nx; k++, jj++)
     {
       t = bz*(u[jj-n2] + u[jj+n2]) +
	   by*(u[jj-n1] + u[jj+n1]) +
	   bx*(u[jj- 1] + u[jj +1]);

       f[jj] = diag[jj] * u[jj] + t;
     }
     f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2])/2. +
	   			by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj -1];
   }

//				Last row in plain

   jj=idx(i,ny,0);
   f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2]) /2.+
	   			by*u[jj-n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(i,ny,1);  jj<idx(i,ny,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2])/2. +
	   			by*u[jj-n1] +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(i,ny,nx);
   f[jj] = diag[jj]*u[jj] + 	bz*(u[jj-n2] + u[jj+n2])/2. +
	   			by*u[jj-n1]/2. +
	   			bx*u[jj -1]/2.;
 }


//	#	#	#	#	#	#	#	#	#
//	Last plain

 i=nz;
//   				First row in plain
   jj=idx(i,0,0);
   f[jj] = diag[jj]*u[jj] + 	bz*u[jj-n2]/2. +
	   			by*u[jj+n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(i,0,1);  jj<idx(i,0,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj-n2]/2. +
	   			by*u[jj+n1]/2. +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(i,0,nx);
   f[jj] = diag[jj]*u[jj] + 	bz*u[jj-n2]/2.  +
	   			by*u[jj+n1]/2. +
	   			bx*u[jj -1]/2.;

//				Inner rows in plain
   for (j=1; j < ny; j++)
   {
     jj = idx(i,j,0);
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj-n2]/2. +
	   			by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj +1]/2.;
     jj++;
     for (k=1; k < nx; k++, jj++)
     {
       t = bz*u[jj-n2] +
	   by*(u[jj-n1] + u[jj+n1])/2. +
	   bx*(u[jj- 1] + u[jj +1])/2.;

       f[jj] = diag[jj] * u[jj] + t;
     }
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj-n2]/2. +
	   			by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj -1]/2.;
   }

//				Last row in plain

   jj=idx(i,ny,0);
   f[jj] = diag[jj]*u[jj] + 	bz*u[jj-n2]/2. +
	   			by*u[jj-n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(i,ny,1);  jj<idx(i,ny,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] + 	bz*u[jj-n2]/2. +
	   		      	by*u[jj-n1]/2. +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(i,ny,nx);
   f[jj] = diag[jj]*u[jj] + bz*u[jj-n2]/2. +
	   		    by*u[jj-n1]/2. +
	   		    bx*u[jj -1]/2.;

 return;
}

//---------------------------------------------------------------------------
//	Derive matrix diagonal for the 7 point stencil
//		and Neumann boundary conditions

void DiagLaplaceEqui3DNeumann :: DeriveDiag(const double *cc)
{
 double cx,cy,cz;
 int i,j,k,ijk;
 int n=VectorLength();

 if (diag==0)   NewGridVector(diag);
// VDcopyC( n, diag, -(bx+by+bz) );

 ijk = 0;
 for ( i=0; i<=nz; i++ )
  {
   for ( j=0; j<=ny; j++ )
    {
     for ( k=0; k<=nx; k++, ijk++ )
      {
	cx = bx; cy = by; cz = bz;
	if ( i==0 || i == nz )  { cx = bx/2.; cy = by/2.; }
	if ( j==0 || j == ny )  { cx = bx/2.; cz = bz/2.; }
	if ( k==0 || k == nx )  { cy = by/2.; cz = bz/2.; }
	diag[ijk] = -(cx+cy+cz);
	if ( k != 0 && k != nx ) diag[ijk] -= cx;
	if ( j != 0 && j != ny ) diag[ijk] -= cy;
	if ( i != 0 && i != nz ) diag[ijk] -= cz;
     }
    }
  }

 if  (cc !=0 ) 			//	Add Helmholtz  term
       VDadd( n, diag, cc );

#ifdef DEBUG_MATRIX
  cout << endl << "Diagonal" << endl;
  PrintVector(n,diag);
#endif
}

//---------------------------------------------------------------------------
//	Derive matrix for the 7 point stencil
//		and Neumann boundary conditions

void DiagLaplaceEqui3DNeumann :: DeriveMatrix()
{
 double cx,cy,cz;
 int i,j,k,n1,n2,ijk,id;
 int nn = VectorLength();
 unsigned long int nn2 = nn*nn;

 n1 = nx+1;  n2 = (nx+1)*(ny+1);

//	Allocate memory for fully occupied matrix if called the first time
//			(Assumes no changes in dimensions!!)
 if (mat==0)   mat = new double[nn2];
 VDcopyC(nn2, mat, 0.0);	// set matrix to Zero

 for ( i=0; i<=nz; i++ )
  {
   for ( j=0; j<=ny; j++ )
    {
     for ( k=0; k<=nx; k++ )
      {
	cx = bx; cy = by; cz = bz;
	if ( i==0 || i == nz )  { cx = bx/2.; cy = by/2.; }
	if ( j==0 || j == ny )  { cx = bx/2.; cz = bz/2.; }
	if ( k==0 || k == nx )  { cy = by/2.; cz = bz/2.; }

	ijk = idx(i,j, k);
	id  = ijk*nn + ijk;
	mat[id] = diag[ijk];
	if ( k != 0  ) mat[id-1 ] = cx;
	if ( k != nx ) mat[id+1 ] = cx;
	if ( j != 0  ) mat[id-n1] = cy;
 	if ( j != ny ) mat[id+n1] = cy;
	if ( i != 0  ) mat[id-n2] = cz;
 	if ( i != nz ) mat[id+n2] = cz;
     }
    }
  }
#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix" << endl;
  PrintMatrix(nn,mat);
#endif
}


//---------------------------------------------------------------------------

void DiagLaplaceEqui3DNeumann :: Defect(const double *f, const double *u,
					double *def) const
{
 if (def==0)
 {
   cout << endl << "DiagLaplaceEqui3DNeumann :: Defect    def-vector is not allocated!!" << endl;
   exit(-1);
 }
 else
 {
//DB   cout << "Defect" << endl;
   Mult(def,u);
   VDaXpY(VectorLength(),def,-1.0,f);
 }
}
//---------------------------------------------------------------------------

void DiagLaplaceEqui3DNeumann :: DirectSolve(double *u, const double *f) const
{
 Equi3D::DirectSolve(u,f);
}

//---------------------------------------------------------------------------

double DiagLaplaceEqui3DNeumann :: DefectNorm(const double *u, const double *f) const
{
  double *aux,s;
//DB   cout << "Defectnorm" << endl;
  NewGridVector(aux);
  Defect(f,u,aux);
  s = VDnorm(VectorLength(),aux);
  DeleteGridVector(aux);
  return sqrt(s);
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3DNeumann :: DefRes(const double *fh, const double *uh,
					double *DH) const
{
  double *aux;
  NewGridVector(aux);
  Defect(fh,uh,aux);
  Restrict(aux, DH);
  DeleteGridVector(aux);
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui3DNeumann :: IntAdd(double *uh, const double *WH) const
{
  double 	t,t2,t4,t8;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const Equi3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = nx+1;   nf2 = (nx+1)*(ny+1);
  nc1 = ncx+1;  nc2 = (ncx+1)*(ncy+1);

  i = 0;
  for (ii=0; ii <= ncz; ii++)
  {
    j = 0;
    for (jj=0; jj <= ncy; jj++)
    {
      k = 0;
      ijk = idx(nf2,nf1, i ,j , k);
      IJK = idx(nc2,nc1, ii,jj, 0);
      for (kk=0; kk <= ncx; kk++, IJK++)
      {
	t  = WH[IJK];
	t2 = t/2.;
	t4 = t/4.;
	t8 = t/8.;

	uh[ijk]	+= t;

	if (i !=  0 ) uh[ijk-nf2]	+= t2;
	if (i != nz ) uh[ijk+nf2]	+= t2;
	if (j !=  0 ) uh[ijk-nf1]	+= t2;
	if (j != ny ) uh[ijk+nf1]	+= t2;
	if (k !=  0 ) uh[ijk-1]		+= t2;
	if (k != nx ) uh[ijk+1]		+= t2;

	if (i !=  0 && j !=  0 ) uh[ijk-nf2-nf1]	+= t4;
	if (i !=  0 && j != ny ) uh[ijk-nf2+nf1]	+= t4;
	if (i != nz && j !=  0 ) uh[ijk+nf2-nf1]	+= t4;
	if (i != nz && j != ny ) uh[ijk+nf2+nf1]	+= t4;
	if (j !=  0 && k !=  0 ) uh[ijk-nf1-1]		+= t4;
	if (j !=  0 && k != nx ) uh[ijk-nf1+1]		+= t4;
	if (j != ny && k !=  0 ) uh[ijk+nf1-1]		+= t4;
	if (j != ny && k != nx ) uh[ijk+nf1+1]		+= t4;
	if (i !=  0 && k !=  0 ) uh[ijk-nf2-1]		+= t4;
	if (i !=  0 && k != nx ) uh[ijk-nf2+1]		+= t4;
	if (i != nz && k !=  0 ) uh[ijk+nf2-1]		+= t4;
	if (i != nz && k != nx ) uh[ijk+nf2+1]		+= t4;

	if (i != 0)
	{
	  if (j != 0)
	  {
	    if (k !=  0) uh[ijk-nf2-nf1-1]	+= t8;
	    if (k != nx) uh[ijk-nf2-nf1+1]	+= t8;
	  }
	  if (j != ny)
	  {
	    if (k !=  0) uh[ijk-nf2+nf1-1]	+= t8;
	    if (k != nx) uh[ijk-nf2+nf1+1]	+= t8;
	  }
	}
	if (i != nz)
	{
	  if (j != 0)
	  {
	    if (k !=  0) uh[ijk+nf2-nf1-1]	+= t8;
	    if (k != nx) uh[ijk+nf2-nf1+1]	+= t8;
	  }
	  if (j != ny)
	  {
	    if (k !=  0) uh[ijk+nf2+nf1-1]	+= t8;
	    if (k != nx) uh[ijk+nf2+nf1+1]	+= t8;
	 }
	}

        k += 2;  ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }
}

//---------------------------------------------------------------------------
void DiagLaplaceEqui3DNeumann :: Inject(const double *uh, double *UH) const
{

  unsigned int	ncx,ncy,ncz,i,j,ii,jj,kk,ijk,IJK;
  const Grid3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();

  i = 0;
  for (ii=0; ii <= ncz; ii++)
  {
    j = 0;
    for (jj=0; jj <= ncy; jj++)
    {
      ijk = idx(i ,j , 0);
      IJK = IDX(ii,jj, 0);
      for (kk=0; kk <= ncx; kk++, IJK++)
      {
	UH[IJK] = uh[ijk];
	ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }
}
//---------------------------------------------------------------------------
void DiagLaplaceEqui3DNeumann :: Restrict(const double *uh, double *UH) const
{
//#define INJECT
#ifdef INJECT
 Inject(uh,UH);
#else

  double  CW;				//  normal weight
  double 	t,t2,t4,t8,c2,c4,c8;
  unsigned int	nf1,nf2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const Equi3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = nx+1;   nf2 = (nx+1)*(ny+1);

  i = 0;
  for (ii=0; ii <= ncz; ii++)
  {
    j = 0;
    for (jj=0; jj <= ncy; jj++)
    {
      k = 0;
      ijk = idx(i ,j , k);
      IJK = IDX(ii,jj, 0);
      for (kk=0; kk <= ncx; kk++, IJK++)
      {
        t2 = t4 = t8 = 0.0;
        c2 = c4 = c8 = 0.0;

	t  = uh[ijk];

        if (i !=  0) {t2 += uh[ijk-nf2]; c2++;}
        if (i != nz) {t2 += uh[ijk+nf2]; c2++;}
        if (j !=  0) {t2 += uh[ijk-nf1]; c2++;}
        if (j != ny) {t2 += uh[ijk+nf1]; c2++;}
        if (k !=  0) {t2 += uh[ijk-  1]; c2++;}
        if (k != nx) {t2 += uh[ijk+  1]; c2++;}

	t2 = t2/2;

	if (i !=  0 && j !=  0) {t4 += uh[ijk-nf2-nf1]; c4++;}
	if (i !=  0 && j != ny) {t4 += uh[ijk-nf2+nf1]; c4++;}
	if (i != nz && j !=  0) {t4 += uh[ijk+nf2-nf1]; c4++;}
	if (i != nz && j != ny) {t4 += uh[ijk+nf2+nf1]; c4++;}

	if (i !=  0 && k !=  0) {t4 += uh[ijk-nf2-1]; c4++;}
	if (i !=  0 && k != nx) {t4 += uh[ijk-nf2+1]; c4++;}
	if (i != nz && k !=  0) {t4 += uh[ijk+nf2-1]; c4++;}
	if (i != nz && k != nx) {t4 += uh[ijk+nf2+1]; c4++;}

	if (k !=  0 && j !=  0) {t4 += uh[ijk-nf1-1]; c4++;}
	if (k !=  0 && j != ny) {t4 += uh[ijk+nf1-1]; c4++;}
	if (k != nx && j !=  0) {t4 += uh[ijk-nf1+1]; c4++;}
	if (k != nx && j != ny) {t4 += uh[ijk+nf1+1]; c4++;}

	t4 = t4/4;

	if (i !=  0 && j !=  0 && k !=  0)  { t8 += uh[ijk-nf2-nf1-1]; c8++;}
	if (i !=  0 && j !=  0 && k != nx)  { t8 += uh[ijk-nf2-nf1+1]; c8++;}
	if (i !=  0 && j != ny && k !=  0)  { t8 += uh[ijk-nf2+nf1-1]; c8++;}
	if (i !=  0 && j != ny && k != nx)  { t8 += uh[ijk-nf2+nf1+1]; c8++;}
	if (i != nz && j !=  0 && k !=  0)  { t8 += uh[ijk+nf2-nf1-1]; c8++;}
	if (i != nz && j !=  0 && k != nx)  { t8 += uh[ijk+nf2-nf1+1]; c8++;}
	if (i != nz && j != ny && k !=  0)  { t8 += uh[ijk+nf2+nf1-1]; c8++;}
	if (i != nz && j != ny && k != nx)  { t8 += uh[ijk+nf2+nf1+1]; c8++;}

	t8 = t8/8;

	CW = 1.0/(1.0 + c2/2. + c4/4. + c8/8.);
//	cout << CW << endl;
	UH[IJK] = (t+t2+t4+t8) * CW ;

	k += 2;  ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }

#endif
}

//	###################################################################
//			DiagLaplaceEqui3DRobin
//	###################################################################
//	Laplace 3D (+c), equidistant tensor product grid, homogeneoeus Robin BC

DiagLaplaceEqui3DRobin :: DiagLaplaceEqui3DRobin(const int mx, const int my, const int mz, const bool coarse)
 : DiagLaplaceEqui3DNeumann(mx,my,mz,coarse)
{;}

//---------------------------------------------------------------------------

DiagLaplaceEqui3DRobin :: DiagLaplaceEqui3DRobin(const int mx, const bool coarse)
 : DiagLaplaceEqui3DNeumann(mx,coarse)
{;}

//---------------------------------------------------------------------------
DiagLaplaceEqui3DRobin :: DiagLaplaceEqui3DRobin(const DiagLaplaceEqui3DRobin &orig)
 : DiagLaplaceEqui3DNeumann(orig)
{;}
//---------------------------------------------------------------------------

DiagLaplaceEqui3DRobin :: ~DiagLaplaceEqui3DRobin()
{
 ;
}
//---------------------------------------------------------------------------
//	Derive matrix diagonal for the 7 point stencil
//		and Robin boundary conditions

void DiagLaplaceEqui3DRobin :: DeriveDiag(const double *cc)
{
 DiagLaplaceEqui3DNeumann :: DeriveDiag(cc);

 double alpha = 1.0;  		// Factor in -u' + alpha* u = 0
 AddBoundaryConst(diag,alpha);

#ifdef DEBUG_MATRIX
  cout << endl << "Diagonal" << endl;
  PrintVector(VectorLength(),diag);
#endif
}















