#ifndef FILE_LAPLACE
#define FILE_LAPLACE

//	###################################################################
//			DiagLaplaceEqui3D
//	###################################################################
/**	Laplace in 3D (provided for -u'' + cu) 
		on an equidistant tensor product grid 
		using a 7-point stencil from Finite Differences.
		The matrix diagonal is stored as a vector.
		All operations are matrix free.

	@param	diag	vector storing the diagonal entries of the matrix
	@param	bx	the constant off-diagonal matrix entry in x-direction
	@param	by	the constant off-diagonal matrix entry in y-direction
	@param	bz	the constant off-diagonal matrix entry in z-direction
*/

class DiagLaplaceEqui3D : public Equi3D
{
public:
/** 	Parameter constructor
	@param	mx	#intervals in x-direction
	@param	my	#intervals in y-direction
	@param	mz	#intervals in z-direction
	@param	coarse	flag whether this grid is the coarsest grid
*/
  DiagLaplaceEqui3D(const int mx, const int my, const int mz, const bool coarse=false);

/**	Parameter constructor with the same #intervals 
	in all directions (and default constructor)
	@param	mx	#intervals in x,y,z-direction
	@param	coarse	flag whether this grid is the coarsest grid
*/
  DiagLaplaceEqui3D(const int mx=1, const bool coarse=false);

/**	Copy constructor
	@param	orig original grid
*/
  DiagLaplaceEqui3D(const DiagLaplaceEqui3D &orig);
  ///
  virtual ~DiagLaplaceEqui3D();

  ///
  DiagLaplaceEqui3D& operator=(const DiagLaplaceEqui3D & orig);

/**	Constructs a finer grid from the actual grid
	@param	fine	the constructed finer grid
*/
  void Refine(DiagLaplaceEqui3D & fine) const;
  
/**	Reference to the next coarser grid
	@return	next coarser grid
*/
//  virtual 
  const DiagLaplaceEqui3D& CoarseGrid()  const 
  		{return *(DiagLaplaceEqui3D*)coarsegrid_;};

/**	Basis method for Diagonal preconditioning, i.e., 
	the correction is the scaled defect. 
		w <-- D^{-1}* r
	@param	w	resulting correction vector
	@param	r	input defect vector
*/
  virtual void DiagPC(double w[], const double r[]) const 
  	{ VDdiv(VectorLength(), w, r, diag); };

/**	Basis method for matrix vector operation.
		f <-- K* u
	@param	f	resulting vector
	@param	u	input  vector	
*/
  virtual void Mult(double f[], const double u[]) const;

/**	Basis method for defect calculation
		def <-- f - K* u 
	@param	f	input rhs vector
	@param	u	input solution vector	
	@param	def	resulting defect vector
*/
  virtual void Defect(const double f[], const double u[], double def[]) const;

/**	Combined defect calculation and restriction to the next coarser grid.
		DH <-- R * ( fh - K * uh )
	@param	fh	input rhs vector (fine grid)
	@param	uh	input solution vector (fine grid)	
	@param	DH	resulting restricted defect vector (coarse grid)
*/
  virtual void DefRes(const double fh[], const double uh[], double DH[]) const;
  
/**	Direct solving of the system of equations. 
	This is only efficient on coarse grids. 
	u = K^{-1}* f
	@param	f	input rhs vector
	@param	u	resulting solution vector	
*/
  virtual void DirectSolve(double u[], const double f[]) const;
  
/**  	omega-Jacobi smoother. Only one sweep is performed.
	This method is not implemented wrt. efficiency.
	@param	u	input/resulting solution vector
	@param	f	input rhs vector
	@param	def	auxilliary vector for storing the old defect
	@param	omega	relaxation parameter
	@param	iter	#smoothing sweeps (paramater is ignored)
*/
  virtual void Jacobi(double u[], const double f[], double def[], 
  			const double omega=1.0, const int iter=1) const; 

/**	Gauss-Seidel forward smoother. Only one sweep is performed.
	@param	u	input/resulting solution vector
	@param	f	input rhs vector
	@param	iter	#smoothing sweeps (paramater is ignored)
*/
  virtual void GaussSeidelForward(double u[], const double f[], const int iter=1) const; 

/**	Gauss-Seidel backward smoother. Only one sweep is performed.
	@param	u	input/resulting solution vector
	@param	f	input rhs vector
	@param	iter	#smoothing sweeps (paramater is ignored)
*/
  virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;

private:
/**	Performs one ADI sweep in x-direction.
	N O T   implemented in 3D!
	@param	u	input/resulting solution vector
	@param	f	input rhs vector
*/
  void ADIx(double u[], const double f[]) const;

/**	Performs one ADI sweep in y-direction.
	N O T   implemented in 3D!
	@param	u	input/resulting solution vector
	@param	f	input rhs vector
*/  
  void ADIy(double u[], const double f[]) const;

/**	ADI forward smoother
							*/
public:
/**  	ADI forward smoother. Only one sweep is performed.
	N O T   implemented in 3D!
	@param	u	input/resulting solution vector
	@param	f	input rhs vector
	@param	def	auxilliary vector
	@param	iter	#smoothing sweeps (paramater is ignored)
*/
  virtual void ADIForward(double u[], const double f[], 
                          double aux[], const int iter=1) const;
/**  	ADI backward smoother. Only one sweep is performed.
	N O T   implemented in 3D!
	@param	u	input/resulting solution vector
	@param	f	input rhs vector
	@param	def	auxilliary vector
	@param	iter	#smoothing sweeps (paramater is ignored)
*/
  virtual void ADIBackward(double u[], const double f[], 
                           double aux[], const int iter=1) const;

/**	Basis method for defect norm calculation.
	@param	u	input solution vector
	@param	f	input rhs vector
	@return	defect norm.
							*/
  virtual double DefectNorm(const double u[], const double f[]) const;  

/**	Initialization of the operator specific data (const. diagonal).
		The method is used in the constructors, etc..	
	If vector cc is set then it will be used for Helmholtz terms
	@param	coarse	Is this the coarse grid 
			(true: factorization for the direct solver)
	@param	cc	vector for the Helmholtz term (=0 if not available) 
*/
  void InitOperator(const bool coarse, const double *cc = 0);

protected:  

//private:
/**	Allocates and calculates the fully occupied square matrix. 
	This is needed for the direct coarse grid solver.
*/  
  virtual void DeriveMatrix(); 

/**	Allocates and calculates the matrix diagonal
	@param	cc	vector of the Helmholtz term (=0 if not available)
*/
  virtual void DeriveDiag(const double *cc); 

/**	Factors the triangular matrices used in the ADI sweeps.
	N O T   implemented in 3D!
*/  
  void FactorADIMatrices();
       
//private:

  double *diag;			// const. diagonal entries of matrix
  double bx,by,bz;		// const. off-diagonal entries of matrix
};


//	###################################################################
//			LaplaceNonEqui3D
//	###################################################################
/**	Laplace in 3D (provided for -u'') 
		on an non-equidistant tensor product grid 
		using a 7-point stencil from Finite Differences.
		The matrix diagonal is stored as a vector. 
		The  off-diagonals are stored as smaller vectors. 
		All operations are matrix free. 
							*/

class LaplaceNonEqui3D : public NonEqui3D
{
public:
/** 	Parameter constructor having numbers of nodes and 
	node coordinates as parameters for all three directions.
	hx[i] := x[i] - x[i-1] with hx[1:nnx-1] and x[0:nnx-1] 
	(hx is stored as hx[0:nnx]). 
							*/
  LaplaceNonEqui3D(const int nnx, const double x[], 
                   const int nny, const double y[], 
	           const int nnz, const double z[], const bool coarse=false);

/**	Parameter constructor with the same #nodes  
	in all directions (and default constructor).	*/
  LaplaceNonEqui3D(const int nnx=1, const double *x = 0, const bool coarse=false);

/** 	Parameter constructor having numbers of intervals and 
	interval lengths as parameters for all three directions.
	xh[i] := x[i] - x[i-1] and 
	xh[0:intx-1] is mapped onto hhx[1:intx] 
	with  hhx[0] = hx[intx+1] = 0. 
							*/
  LaplaceNonEqui3D(const int intx, const int inty, const int intz, 
            const double xh[], const double yh[], const double zh[], 
	    const bool coarse=false);

/**	Copy constructor
							*/
  LaplaceNonEqui3D(const LaplaceNonEqui3D &orig);
  ///
  virtual ~LaplaceNonEqui3D();

  ///
  LaplaceNonEqui3D& operator=(const LaplaceNonEqui3D & orig);

/**	Construct a finer grid from the given grid in the class.
							*/
  void Refine(LaplaceNonEqui3D & fine) const;
  
/**	Reference to the next coarser grid
							*/
//  virtual 
  const LaplaceNonEqui3D& CoarseGrid()  const 
  		{return *(LaplaceNonEqui3D*)coarsegrid_;};

/**	Basis method for Diagonal preconditioning
		w <-- D^{-1}* r 			*/
  void DiagPC(double *w, const double *r) const 
  	{VDdiv(VectorLength(), w, r, diag);};

/**	Basis method for matrix vector operation
		f <-- K* u 			*/
  virtual void Mult(double *f, const double *u) const;

/**	Basis method for defect calculation
		def <-- f - K* u 			*/
  virtual void Defect(const double *f, const double *u, double *def) const;

/**	Combined defect calculation and restriction
		DH <-- R * ( fh - K * uh )		*/
  virtual void DefRes(const double *fh, const double *uh, double *DH) const;
  
/**	Direct solving of (coarse) system
							*/	
  virtual void DirectSolve(double *u, const double *f) const;
  
/**  	Jacobi smoother
							*/
  virtual void Jacobi(double *u, const double *f, double *def, 
  			const double omega=1.0, const int iter=1) const; 

/**	Gauss-Seidel forward smoother
							*/
  virtual void GaussSeidelForward(double *u, const double *f, const int iter=1) const; 

/**	Gauss-Seidel backward smoother
							*/
  virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;

private:
  void ADIx(double u[], const double f[]) const;
  void ADIy(double u[], const double f[]) const;

/**	ADI forward smoother
							*/
public:
  virtual void ADIForward(double *u, const double *f, 
                          double *aux, const int iter=1) const;
/**	ADI backward smoother
							*/
  virtual void ADIBackward(double *u, const double *f, 
                           double *aux, const int iter=1) const;

/**	Virtual basis method for defect norm calculation
							*/
  virtual double DefectNorm(const double *u, const double *f) const;  

/**	Initialization of the operator specific data (const. diagonal).
		The method is used in the constructors, etc..	
	If vector cc is set then it will be used for Helmholtz terms
		*/
  void InitOperator(const bool coarse, const double *cc = 0);

protected:  

private:
/**	Allocates and calculates the fully occupied square matrix. 
								*/  
  void DeriveMatrix(); 

  void OffDiag1D(const int n, const double h[], double bl[], double br[]);
  
       
private:

  double *diag;			// diagonal entries of matrix
  double *bxr, *bxl;		// off-diagonal entries of matrix in x-direction
  double *byr, *byl;
  double *bzr, *bzl;
};

//	###################################################################
//			DiagLaplaceEqui3Neumann
//	###################################################################
/**	Laplace in 3D (provided for -u'' + cu)
		on an equidistant tensor product grid
		using a 7-point stencil from Finite Differences.
	Homogeneous Neumann boundary conditions
		The matrix diagonal is stored as a vector.
		All operations are matrix free. 
							*/

class DiagLaplaceEqui3DNeumann : public DiagLaplaceEqui3D
{
public:
/** 	Parameter constructor
							*/
  DiagLaplaceEqui3DNeumann(const int mx, const int my, const int mz, const bool coarse=false);

/**	Parameter constructor with the same #intervals
	in all directions (and default constructor).	*/
  DiagLaplaceEqui3DNeumann(const int mx=1, const bool coarse=false);

/**	Copy constructor
							*/
  DiagLaplaceEqui3DNeumann(const DiagLaplaceEqui3DNeumann &orig);
  ///
  virtual ~DiagLaplaceEqui3DNeumann();

/**	Construct a finer grid from the given grid in the class.
							*/
  void Refine(DiagLaplaceEqui3DNeumann & fine) const;

/**	Reference to the next coarser grid
							*/
//  virtual
  const DiagLaplaceEqui3DNeumann& CoarseGrid()  const
  		{return *(DiagLaplaceEqui3DNeumann*)coarsegrid_;};

/**	Basis method for Diagonal preconditioning
		w <-- D^{-1}* r 			*/
  void DiagPC(double *w, const double *r) const
  	{VDdiv(VectorLength(), w, r, diag);};

/**	Basis method for matrix vector operation
		f <-- K* u 			*/
  virtual void Mult(double *f, const double *u) const;
  void Mult_long(double *f, const double *u) const;

 /**	Basis method for defect calculation
 		def <-- f - K* u 			*/
  virtual void Defect(const double *f, const double *u, double *def) const;

 /**	Combined defect calculation and restriction
 		DH <-- R * ( fh - K * uh )		*/
   virtual void DefRes(const double *fh, const double *uh, double *DH) const;

 /**	Direct solving of (coarse) system
 							*/
   virtual void DirectSolve(double *u, const double *f) const;

 /**  	Jacobi smoother
 							*/
//   virtual void Jacobi(double *u, const double *f, double *def,
//   			const double omega=1.0, const int iter=1) const;

 /**	Gauss-Seidel forward smoother
 							*/
//   virtual void GaussSeidelForward(double *u, const double *f, const int iter=1) const;

 /**	Gauss-Seidel backward smoother
 							*/
//   virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;

   void Inject(const double *uh, double *UH) const;
   void Restrict(const double *uh, double *UH) const;
   void IntAdd(double *uh, const double *WH) const;


private:
//   void ADIx(double u[], const double f[]) const;
//   void ADIy(double u[], const double f[]) const;

public:

/**	ADI forward smoother
							*/
//  virtual void ADIForward(double *u, const double *f, 
//                          double *aux, const int iter=1) const;
/**	ADI backward smoother
							*/
//  virtual void ADIBackward(double *u, const double *f,
//                           double *aux, const int iter=1) const;

/**	Virtual basis method for defect norm calculation
							*/
  virtual double DefectNorm(const double *u, const double *f) const;  

/**	Initialization of the operator specific data (const. diagonal).
  	      The method is used in the constructors, etc..
      If vector cc is set then it will be used for Helmholtz terms
 							*/
//  void InitOperator(const bool coarse, const double *cc = 0); 

protected:

/**	Allocates and calculates the fully occupied square matrix.
								*/
  virtual void DeriveMatrix();

/**	Allocates and calculates the matrix diagonal
								*/
  virtual void DeriveDiag(const double *cc);
//  void FactorADIMatrices();

};

//	###################################################################
//			DiagLaplaceEqui3Robin
//	###################################################################
/**	Laplace in 3D (provided for -u'' + cu)
		on an equidistant tensor product grid
		using a 7-point stencil from Finite Differences.
	Homogeneous Robin boundary conditions: -u' + 1*u = 0
		The matrix diagonal is stored as a vector.
		All operations are matrix free. 
							*/

class DiagLaplaceEqui3DRobin : public DiagLaplaceEqui3DNeumann
{
public:
/** 	Parameter constructor
							*/
  DiagLaplaceEqui3DRobin(const int mx, const int my, const int mz, const bool coarse=false);

/**	Parameter constructor with the same #intervals
	in all directions (and default constructor).	*/
  DiagLaplaceEqui3DRobin(const int mx=1, const bool coarse=false);

/**	Copy constructor
							*/
  DiagLaplaceEqui3DRobin(const DiagLaplaceEqui3DRobin &orig);
  
  ///
  virtual ~DiagLaplaceEqui3DRobin();

/**	Reference to the next coarser grid
							*/
//  virtual
  const DiagLaplaceEqui3DRobin& CoarseGrid()  const
  		{return *(DiagLaplaceEqui3DRobin*)coarsegrid_;};

protected:

/**	Allocates and calculates the fully occupied square matrix.
								*/
//  virtual void DeriveMatrix();

/**	Allocates and calculates the matrix diagonal
								*/
  virtual void DeriveDiag(const double *cc);

};


#endif // FILE_LAPLCE
