#ifndef FILE_LAPLACEVARY
#define FILE_LAPLACEVARY

//	###################################################################
//			LaplaceVaryNonEqui3D
//	###################################################################
/**	Laplace in 3D (provided for -(a*u')') with homog. Dirichlet BC
		on an non-equidistant tensor product grid 
		using a 7-point stencil from Finite Differences.
		The matrix diagonal is stored as a vector. 
		The  tensor product parts of the off-diagonals are stored 
		     as smaller vectors. 
		The coefficent dependent parts of the off-diagonals are always 
		    calculated from the coefficent vector ak.
		All operations are matrix free. 
							*/

class LaplaceVaryNonEqui3D : public NonEqui3D
{
public:
/** 	Parameter constructor having numbers of nodes and 
	node coordinates as parameters for all three directions.
	hx[i] := x[i] - x[i-1] with hx[1:nnx-1] and x[0:nnx-1] 
	(hx is stored as hx[0:nnx]). 
							*/
  LaplaceVaryNonEqui3D(const int nnx, const double x[], 
                   const int nny, const double y[], 
	           const int nnz, const double z[], const bool coarse=false);

/**	Parameter constructor with the same #nodes  
	in all directions (and default constructor).	*/
  LaplaceVaryNonEqui3D(const int nnx=1, const double *x = 0, const bool coarse=false);

/** 	Parameter constructor having numbers of intervals and 
	interval lengths as parameters for all three directions.
	xh[i] := x[i] - x[i-1] and 
	xh[0:intx-1] is mapped onto hhx[1:intx] 
	with  hhx[0] = hx[intx+1] = 0. 
							*/
  LaplaceVaryNonEqui3D(const int intx, const int inty, const int intz, 
            const double xh[], const double yh[], const double zh[], 
	    const bool coarse=false);

/**  Parameter constructor to setup the operator only from grid info. 
*/
//   LaplaceVaryNonEqui3D(const NonEqui3D &orig);

/**	Copy constructor
							*/
  LaplaceVaryNonEqui3D(const LaplaceVaryNonEqui3D &orig);
  ///
  virtual ~LaplaceVaryNonEqui3D();

  ///
  LaplaceVaryNonEqui3D& operator=(const LaplaceVaryNonEqui3D & orig);

/**	Construct a finer grid from the given grid in the class.
							*/
  void Refine(LaplaceVaryNonEqui3D & fine) const;
  
/**	Reference to the next coarser grid
							*/
//  virtual 
  const LaplaceVaryNonEqui3D& CoarseGrid()  const 
  		{return *(LaplaceVaryNonEqui3D*)coarsegrid_;};

/**	Basis method for Diagonal preconditioning
		w <-- D^{-1}* r 			*/
  void DiagPC(double *w, const double *r) const 
  	{VDdiv(VectorLength(), w, r, diag);};

/**	Basis method for matrix vector operation
		f <-- K* u 			*/
  virtual void Mult(double *f, const double *u) const;

/**	Basis method for defect calculation
		def <-- f - K* u 			*/
  virtual void Defect(const double *f, const double *u, double *def) const;

/**	Combined defect calculation and restriction
		DH <-- R * ( fh - K * uh )		*/
  virtual void DefRes(const double *fh, const double *uh, double *DH) const;
  virtual void DefRes_fast(const double *fh, const double *uh, double *DH) const;
 
/**	Direct solving of (coarse) system
							*/	
  virtual void DirectSolve(double *u, const double *f) const;
  
/**  	Jacobi smoother
							*/
  virtual void Jacobi(double *u, const double *f, double *def, 
  			const double omega=1.0, const int iter=1) const; 

/**	Gauss-Seidel forward smoother
							*/
  virtual void GaussSeidelForward(double *u, const double *f, const int iter=1) const; 

/**	Gauss-Seidel backward smoother
							*/
  virtual void GaussSeidelBackward(double *u, const double *f, const int iter=1) const;

private:
  void ADIx(double u[], const double f[]) const;
  void ADIy(double u[], const double f[]) const;

/**	ADI forward smoother
							*/
public:
  virtual void ADIForward(double *u, const double *f, 
                          double *aux, const int iter=1) const;
/**	ADI backward smoother
							*/
  virtual void ADIBackward(double *u, const double *f, 
                           double *aux, const int iter=1) const;

/**	Virtual basis method for defect norm calculation
							*/
  virtual double DefectNorm(const double *u, const double *f) const;  

/**	Initialization of the operator specific data (const. diagonal).
		The method is used in the constructors, etc..	
	If vector cc is set then it will be used for Helmholtz terms
		*/
  void InitOperator(const bool coarse, const double *cc = 0);

protected:  

private:
/**	Allocates and calculates the fully occupied square matrix. 
								*/  
  void DeriveMatrix(); 

  void OffDiag1D(const int n, const double h[], double bl[], double br[]);
  
       
private:

  double *diag;			// diagonal entries of matrix
  double *ak;                 // diffusion coefficient 
  double *bxr, *bxl;		// off-diagonal entry parts of matrix in x-direction
  double *byr, *byl;
  double *bzr, *bzl;
};
#endif
