#ifndef FILE_MGPROCS
#define FILE_MGPROCS

#include "grid.hpp"

/**	CtrlRec contains all the parameters to control the multigrid iteration
@param iterat	maximal number of multigrid iterations
@param zykluswahl	V- or W-cycle ( 1 / 2)
@param vGamma		number of pre-smoothing sweeps
@param nGamma		number of post-smoothing sweeps
@param res		type of restriction (injection/linear/bilinear = 0/1/2)
@param inter		type of interpolation (linear/bilinear = 1/2)
			only bi-/trilinear is implemented
@param vGlatt		type of pre-smoothing
			(omega_Jacobi/GS_forw/GS_backw/ADI_forw/ADI_backw = 1/2/3/4/5)
@param nGlatt		type of post-smoothing
@param AllgV		general V-cycle (T/F)?
@param Full		Full multigrid approach (T/F)?
@param eps		stopping criteria for the multigrid iteration (defect norm)
*/
struct CtrlRec
 {
   short int        iterat, zykluswahl, vGamma, nGamma;
   GridtransferType res, inter;
   SmoothingType    vGlatt, nGlatt;
   bool             AllgV, Full;
   double           eps;
 };

//####################################################################
//		Should be implemented as class
//####################################################################

/**	VectorRec contains the vectors for
	solution, right hand side, defect on one grid.
@param	u	solution vector (= calculated coarse grid correction)
@param  f	right hand side (= input coarse defect)
@param  def	defect vector, that stores the defect after presmoothing
*/
struct VectorRec
 {
   double *u, *f, *def;
 };

//####################################################################
//		Should be also adapted to class MultiGrid
//####################################################################

/**	Recursive multigrid procedure on one level.
	Calls itselfs on the next coarser level.
	@param	grid	fine grid
	@param	f_data	pointer at fine grid vectors
	@param	f_ctrl	pointer at fine grid control structure
*/
void MGM(const Grid &grid, VectorRec *f_data, CtrlRec *f_ctrl);

/**	Multigrid iteration with norm calculatation for stopping criteria.
	@param	nlevels		number of levels
	@param	gf		finest grid
	@param	vecs		array of data (=vectors) for the grids
	@param	ctrl		array of control structures for the grids
	@param	iter		resulting number of iterations
	@param	norm		relative error reduction in the defect norm
*/
void MultigridIteration(unsigned int nlevels,
	const Grid &gf,VectorRec *vecs, CtrlRec *ctrl,
	int &iter, double  &norm);

/**	Multigrid iteration without norm calculatation for stopping criteria,
	i.e., the only stopping criteria is the maximal number of iterations
	in ctrl[nlevels-1].iterat
	@see MultigridIteration
*/
void MultigridIterationNoNorm(unsigned int nlevels,
	const Grid &gf,VectorRec *vecs, CtrlRec *ctrl,
	int &iter);

/**	Derivation of the control structure for the coarser grid.
	@param	Grob	resulting coarse grid control structure
	@param	Fein	fine grid control structure
*/
void GrobSteuer(CtrlRec &Grob, const CtrlRec &Fein);


#endif	// MGPROCS
