#ifndef FILE_MULTIGRID
#define FILE_MULTIGRID
#include<cassert>
#include "grid.hpp"
#include "grid2d.hpp"
#include "grid3d.hpp"
#include "laplace2d.hpp"
#include "laplace3d.hpp"
#include "mgprocs.hpp"


//####################################################################
//		Template class MultiGrid
//####################################################################

/**	The class MultiGrid contains the multigrid algorithm for the
	grid and operator provided by the template class GRID.
	@short	Realization of multigrid algorithms
*/
template<class GRID>
class MultiGrid
{
public:
 /**	Initializes Multigrid with all grid hierarchies depending on operator and
 	grid defined in the template GRID.
 	The vectors for solution and right hand side of the finest level have to be
 	allocated externally.
	@param	simple		controls the multigrid algorithm
	@param	levels		number of grids and number of finest level used
	@param	grid_coarse	coarsest grid to build the grid hierarchy from
 */
  MultiGrid(const CtrlRec& simple, const unsigned int levels,
  	    const GRID& grid_coarse);

 /**	Initializes Multigrid with all grid hierarchies depending on operator and
 	grid defined in the template GRID.
 	The vectors for solution and right hand side of the finest level are
 	allocated internally. This constructor should be used for Full Multigrid techniques.
	@param	simple		controls the multigrid algorithm
	@param	levels		number of finest level used
	@param	grid_coarse	coarsest grid to build the grid hierarchy from
	@param	alloclevels	number of allocated levels (may be largen then levels)
 */
  MultiGrid(const CtrlRec& simple, const unsigned int levels,
  	    const GRID& grid_coarse, const unsigned int alloclevels);

//	Declaration without definition for copy constructor and assignment operator
  explicit MultiGrid(const MultiGrid<GRID> &orig);
private:
  MultiGrid<GRID> & operator=(const MultiGrid<GRID> &orig);
public:


  ///
  ~MultiGrid();

/**	Adds a vector to the diagonals of all grids, i.e.,
	a Helmholtz term is added to the pde. The input vector is injected onto the
	coarser grids.
	If cc==NULL then the already initialized Laplace operatopr is used,
	else the fine grid data in cc[] are used to update the operator with
	the Helmholtz term  +cu .

	@param	cc		Helmholtz term values for all grid points
*/
  void AddDiagTerm(const double *cc = 0);

/**	Solves the system of equations  K*u = f .
	The matrix K, all grids and aux. data are already initialized.
	@param	u	solution vector
	@param  f	right hand side
	@param	iter	number of mg iterations to solve
	@param	norm	relative error reduction in the defect norm
	@param	bnorm	should the norms are calculated (speed)
	@see AddDiagTerm
*/
  void Solve(double u[], double f[], int &iter, double &norm,
  		const bool bnorm=true);

/**	Injects a vector of the actual fine grid to the next coarser grid.
		( UH <-- R * uh )
	@param	uh	fine grid vector
	@param	UH	coarse grid vector
*/
  void Inject(const double *uh, double *UH) const
			{ grids[ GetCalcLevel()-1 ].Inject(uh,UH); };

/**	Restricts (bi-/tri-linear) a vector of the actual fine grid to the next coarser grid.
		( UH <-- R * uh )
	@param	uh	fine grid vector
	@param	UH	coarse grid vector
*/

  void Restrict(const double *uh, double *UH) const
			{ grids[ GetCalcLevel()-1 ].Restrict(uh,UH); };

/**	Interpolates (bi-/tri-linear) a vector from the next coarser grid to the actual fine grid
		( wh <-- P * WH )
	@param	wh	fine grid vector
	@param	WH	coarse grid vector
*/
  void Interpolate(double *wh, const double *WH) const
			{ grids[ GetCalcLevel()-1 ].Interpolate(wh,WH); };

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/**	Set the actual finest level for calculations
	@param	level	actual finest level
*/
  void SetCalcLevel(const unsigned int level);

/**
	@return		a reference to the actual finest grid
*/
  const GRID& GetFineGrid() const { return grids[nlevels-1];};

/**
	@param	lev	number of thel level (0..nlevels-1)
	@return		a reference to a grid in thehierachy
*/
  const GRID& GetGrid(const int lev) const
  	{ assert(0<=lev && lev<nlevels); return grids[lev];};

/**
	@return		the actual finest level for calculations
*/
  unsigned int GetCalcLevel() const {return nlevels;};

/**
	@return		 the finest level allocated for calculations
*/
  unsigned int NumAllocatedLevels() const {return highestlevel;};

/**
	@return		 the length of a vector of the actual finest grid.
	@see	SetCalcLevel
*/
  int VectorLength() const {return GetFineGrid().VectorLength();};

/**
	@return	 the number of points in x-directions of the actual finest grid.
*/
  int GetIntX() const {return GetFineGrid().GetIntX();};

/**	Preconditioning with the diagonal of the operator for the actual finest grid:
		w := D^{-1} r
	@param	w	resulting preconditioned correction vector
	@param	r	residuum vector
	@see SetCalcLevel AddDiagTerm
*/
  void DiagPC(double w[], const double r[]) const {GetFineGrid().DiagPC(w,r);};
/**	Multiplication with the operator of the actual finest grid:
		w := D^{-1} * r
	@param	r	input vector
	@param	w	output vector
	@see SetCalcLevel AddDiagTerm
*/
  void Mult(double f[], const double u[]) const {GetFineGrid().Mult(f,u);};

/**
	@return 	space dimension of the problem
*/
  int  NumberDimensions() const {return GetFineGrid().NumberDimensions();};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

private:
/**	Nearly the same method as the next ini_level method. The difference
	consists in static binding of u and f.
*/
  void ini_level(const GRID &grid_l, VectorRec &vecs_l, const CtrlRec &ctrl_l,
		 double *u_in=0, double *f_in=0);

/**	Allocates memory for the internal arrays u, f, def on the level described
	in the control record. The defect vector is allocated only if Jacobi smoothing
	is used in pre- or post-smoothing. Vectors u, f, are allocated if the
	appropriate boolean variables are false. If they are true (usually on the
	finest grid) then u, f  are dynamically binded.
	@param	grid_l		recent grid
	@param	vecs_l		structure containing vectors of recent grid
	@param	ctrl_l		control structure of recent grid
	@param	bu		is solution externally allocated?
	@param	bf		is rhs externally allocated?
*/
  void ini_level(const GRID &grid_l, VectorRec &vecs_l, const CtrlRec &ctrl_l,
		 const bool bu, const bool bf);

private:
   unsigned int nlevels;	// actual highest level for working (nlevels<=highestlevel)
   unsigned int highestlevel;	// highest level where memory is allocated
   bool         u_extern; 	// are the vectors u and f on the fine grid
   bool         f_extern;	//     externally  allocated?
   GRID	    *grids;		// array of grids
   CtrlRec      *ctrl;		// array of control structures
   VectorRec    *vecs;		// array of structures for vectors
};



#include "multigrid.cpp"

#endif	// MULTIGRID
